"""
 @file
 @brief This file contains unit tests for selected main window flows
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import importlib
import os
import sys
import tempfile
import threading
import types
import unittest
import zipfile
from contextlib import ExitStack
from datetime import datetime, timedelta
from unittest.mock import patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from PyQt5.QtCore import QCoreApplication, Qt
from PyQt5.QtWidgets import QApplication

from classes import info
from classes.project_data import ProjectDataStore
from classes.updates import UpdateManager
from qt_test_app import ensure_app_state as ensure_qt_app_state, get_or_create_app

QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)


class DummySettings:
    actionType = types.SimpleNamespace(LOAD="load")

    def __init__(self):
        self.values = {
            "recent_projects": [],
            "history-limit": 20,
            "recovery-limit": 10,
            "default-profile": "HD 720p 30 fps",
            "default-samplerate": 48000,
            "default-channels": 2,
        }
        self.default_paths = {}

    def get(self, key):
        return self.values.get(key)

    def set(self, key, value):
        self.values[key] = value

    def setDefaultPath(self, action, path):
        self.default_paths[action] = path


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()
        self.project = None
        self.updates = None
        self.window = None
        self.logger_libopenshot = None

    def get_settings(self):
        return self.settings

    def _tr(self, text):
        return text


def ensure_app_state(app):
    return ensure_qt_app_state(
        app,
        DummySettings,
        project_factory=ProjectDataStore,
        updates_factory=UpdateManager,
        extra_attrs={"window": None, "logger_libopenshot": None},
    )


class SignalRecorder:
    def __init__(self):
        self.calls = []

    def emit(self, *args):
        self.calls.append(args)


class MainWindowTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app, cls._owns_app = get_or_create_app(DummyApp)
        cls.app = ensure_app_state(app)
        cls._web_backend_patcher = patch.object(info, "WEB_BACKEND", "qwidget")
        cls._web_backend_patcher.start()
        metrics = types.ModuleType("classes.metrics")
        metrics.track_metric_session = lambda *args, **kwargs: None
        metrics.track_metric_screen = lambda *args, **kwargs: None
        sys.modules["classes.metrics"] = metrics
        sys.modules.pop("windows.views.timeline", None)
        sys.modules.pop("windows.main_window", None)
        cls.main_window_module = importlib.import_module("windows.main_window")

    @classmethod
    def tearDownClass(cls):
        cls._web_backend_patcher.stop()
        if getattr(cls, "_owns_app", False) and cls.app:
            cls.app.quit()

    def setUp(self):
        ensure_app_state(self.app)
        self.app.settings = DummySettings()
        self.app.window = None

    def tearDown(self):
        ensure_app_state(self.app)

    def test_manage_recovery_files_keeps_daily_and_historical_limits(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            recovery_dir = os.path.join(tmpdir, "recovery")
            os.mkdir(recovery_dir)

            with patch.object(self.main_window_module.info, "RECOVERY_PATH", recovery_dir):
                # Freeze the module clock so the retention logic and fixture
                # timestamps use the same notion of "today".
                now = datetime(2026, 3, 21, 12, 0, 0)

                class FixedDateTime(datetime):
                    @classmethod
                    def now(cls, tz=None):
                        if tz is not None:
                            return tz.fromutc(now.replace(tzinfo=tz))
                        return now

                files = [
                    ("100-newest-project.zip", now),
                    ("090-older-today-project.zip", now - timedelta(hours=1)),
                    ("080-yesterday-project.zip", now - timedelta(days=1)),
                    ("070-two-days-project.zip", now - timedelta(days=2)),
                ]
                for name, dt in files:
                    path = os.path.join(recovery_dir, name)
                    with open(path, "w", encoding="utf-8") as handle:
                        handle.write("x")
                    ts = dt.timestamp()
                    os.utime(path, (ts, ts))

                fake_window = types.SimpleNamespace()
                with patch.object(self.main_window_module, "datetime", FixedDateTime):
                    self.main_window_module.MainWindow.manage_recovery_files(fake_window, 1, 1, "project")

                remaining = sorted(os.listdir(recovery_dir))
                self.assertEqual(remaining, ["080-yesterday-project.zip", "100-newest-project.zip"])

    def test_save_project_emits_saved_signal_on_success(self):
        saved = SignalRecorder()
        failed = SignalRecorder()
        history_calls = []
        save_calls = []

        self.app.project = types.SimpleNamespace(save=save_calls.append)
        self.app.updates = types.SimpleNamespace(
            save_history=lambda project, limit: history_calls.append((project, limit))
        )
        self.app.settings.values["history-limit"] = 42

        fake_window = types.SimpleNamespace(
            lock=threading.Lock(),
            save_recovery=lambda path: save_calls.append(f"recovery:{path}"),
            ProjectSaved=saved,
            ProjectSaveFailed=failed,
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "project.osp")
            self.main_window_module.MainWindow.save_project(fake_window, project_path)

            self.assertEqual(history_calls, [(self.app.project, 42)])
            self.assertEqual(save_calls, [f"recovery:{project_path}", project_path])
            self.assertEqual(saved.calls, [(project_path,)])
            self.assertEqual(failed.calls, [])

    def test_optimized_preview_actions_use_cached_menu_targets_when_selection_is_empty(self):
        proxy_calls = []
        file_obj = types.SimpleNamespace(id="F1", data={"id": "F1", "media_type": "video"})
        fake_window = types.SimpleNamespace(
            selected_files=lambda: [],
            _optimized_preview_target_file_ids=["F1"],
            proxy_service=types.SimpleNamespace(
                remove_for_files=lambda files: proxy_calls.append(("remove", [getattr(f, "id", None) for f in files])),
                create_for_files=lambda files: proxy_calls.append(("create", [getattr(f, "id", None) for f in files])),
                use_existing_for_files=lambda files: proxy_calls.append(("locate", [getattr(f, "id", None) for f in files])),
                cancel_for_files=lambda files: proxy_calls.append(("cancel", [getattr(f, "id", None) for f in files])),
                delete_and_unlink_for_files=lambda files: proxy_calls.append(("delete", [getattr(f, "id", None) for f in files])),
            ),
        )
        fake_window._optimized_preview_files_for_action = lambda: self.main_window_module.MainWindow._optimized_preview_files_for_action(fake_window)

        with patch.object(self.main_window_module.File, "get", return_value=file_obj):
            files = self.main_window_module.MainWindow._optimized_preview_files_for_action(fake_window)
            self.assertEqual([f.id for f in files], ["F1"])

            self.main_window_module.MainWindow.actionOptimizedPreviewRemove_trigger(fake_window)
            self.main_window_module.MainWindow.actionOptimizedPreviewCreate_trigger(fake_window)
            self.main_window_module.MainWindow.actionOptimizedPreviewDeleteAndUnlink_trigger(fake_window)

        self.assertEqual(proxy_calls, [("remove", ["F1"]), ("create", ["F1"]), ("delete", ["F1"])])

    def test_optimized_preview_cancel_targets_current_file_only(self):
        proxy_calls = []
        file_obj = types.SimpleNamespace(id="F1", data={"id": "F1", "media_type": "video"})
        fake_window = types.SimpleNamespace(
            current_file_id=lambda: "F1",
            selected_files=lambda: [
                types.SimpleNamespace(id="F1", data={"id": "F1", "media_type": "video"}),
                types.SimpleNamespace(id="F2", data={"id": "F2", "media_type": "video"}),
                types.SimpleNamespace(id="F3", data={"id": "F3", "media_type": "video"}),
            ],
            _optimized_preview_target_file_ids=["F1", "F2", "F3"],
            proxy_service=types.SimpleNamespace(
                cancel_for_files=lambda files: proxy_calls.append(("cancel", [getattr(f, "id", None) for f in files])),
            ),
        )
        fake_window._optimized_preview_files_for_action = lambda: self.main_window_module.MainWindow._optimized_preview_files_for_action(fake_window)
        fake_window._optimized_preview_file_for_cancel_action = lambda: self.main_window_module.MainWindow._optimized_preview_file_for_cancel_action(fake_window)

        with patch.object(self.main_window_module.File, "get", return_value=file_obj):
            self.main_window_module.MainWindow.actionOptimizedPreviewCancel_trigger(fake_window)

        self.assertEqual(proxy_calls, [("cancel", ["F1"])])

    def test_open_project_missing_file_removes_recent_project_and_seeks_start(self):
        status_messages = []
        removed = []
        loaded_recent = []
        move_calls = []
        restore_cursor = []
        speed_calls = SignalRecorder()
        pause_calls = SignalRecorder()

        player = types.SimpleNamespace(Seek=lambda frame: move_calls.append(("seek", frame)))
        preview_thread = types.SimpleNamespace(player=player)
        video_preview = types.SimpleNamespace(
            clearTransformState=lambda: move_calls.append(("clear_transform",)),
            size=lambda: "preview-size",
        )

        fake_window = types.SimpleNamespace(
            SpeedSignal=speed_calls,
            PauseSignal=pause_calls,
            videoPreview=video_preview,
            clearSelections=lambda: move_calls.append(("clear_selections",)),
            statusBar=types.SimpleNamespace(showMessage=lambda text, ms: status_messages.append((text, ms))),
            remove_recent_project=removed.append,
            load_recent_menu=lambda: loaded_recent.append(True),
            movePlayhead=lambda frame: move_calls.append(("playhead", frame)),
            preview_thread=preview_thread,
            SetWindowTitle=lambda: None,
            refreshFilesSignal=SignalRecorder(),
            refreshFrameSignal=SignalRecorder(),
            MaxSizeChanged=SignalRecorder(),
            actionSave_trigger=lambda: None,
        )

        self.app.project = types.SimpleNamespace(needs_save=lambda: False)
        self.app.updates = types.SimpleNamespace(load_history=lambda project: None)
        self.app.window = fake_window
        self.app.setOverrideCursor = lambda cursor: None
        self.app.restoreOverrideCursor = lambda: restore_cursor.append(True)

        with tempfile.TemporaryDirectory() as tmpdir:
            missing_path = os.path.join(tmpdir, "missing.osp")
            with patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None):
                self.main_window_module.MainWindow.open_project(
                    fake_window,
                    missing_path,
                    clear_thumbnails=True,
                )

            self.assertEqual(removed, [missing_path])
            self.assertEqual(loaded_recent, [True])
            self.assertTrue(status_messages)
            self.assertIn("missing", status_messages[0][0].lower())
            self.assertIn(("seek", 1), move_calls)
            self.assertIn(("playhead", 1), move_calls)
            self.assertEqual(speed_calls.calls, [(0,)])
            self.assertEqual(pause_calls.calls, [()])
            self.assertEqual(restore_cursor, [True])

    def test_save_recovery_creates_zip_and_calls_retention(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "demo.osp")
            with open(project_path, "w", encoding="utf-8") as handle:
                handle.write("project-data")

            recovery_dir = os.path.join(tmpdir, "recovery")
            os.mkdir(recovery_dir)
            self.app.settings.values["recovery-limit"] = 10

            managed = []
            fake_window = types.SimpleNamespace(
                manage_recovery_files=lambda daily, historical, name: managed.append((daily, historical, name))
            )

            with ExitStack() as stack:
                stack.enter_context(
                    patch.object(self.main_window_module.info, "RECOVERY_PATH", recovery_dir)
                )
                stack.enter_context(
                    patch.object(self.main_window_module, "time", lambda: 1234567890)
                )
                self.main_window_module.MainWindow.save_recovery(fake_window, project_path)

            zip_path = os.path.join(recovery_dir, "1234567890-demo.zip")
            self.assertTrue(os.path.exists(zip_path))
            with zipfile.ZipFile(zip_path, "r") as archive:
                self.assertEqual(archive.namelist(), ["demo.osp"])
                self.assertEqual(archive.read("demo.osp").decode("utf-8"), "project-data")
            self.assertEqual(managed, [(7, 3, "demo")])

    def test_close_event_cancel_keeps_window_open(self):
        tutorial_calls = []
        save_calls = []
        event_calls = []

        fake_window = types.SimpleNamespace(
            tutorial_manager=types.SimpleNamespace(
                re_show_dialog=lambda: tutorial_calls.append("reshow"),
            ),
            actionSave_trigger=lambda: save_calls.append("save"),
            shutting_down=False,
        )
        self.app.project = types.SimpleNamespace(needs_save=lambda: True)

        event = types.SimpleNamespace(
            accept=lambda: event_calls.append("accept"),
            ignore=lambda: event_calls.append("ignore"),
        )

        with patch.object(
            self.main_window_module.QMessageBox,
            "question",
            return_value=self.main_window_module.QMessageBox.Cancel,
        ):
            self.main_window_module.MainWindow.closeEvent(fake_window, event)

        self.assertEqual(save_calls, [])
        self.assertEqual(tutorial_calls, ["reshow"])
        self.assertEqual(event_calls, ["ignore"])
        self.assertFalse(fake_window.shutting_down)

    def test_close_event_yes_saves_and_continues_shutdown(self):
        calls = []
        tracker = []
        event_calls = []

        fake_window = types.SimpleNamespace(
            tutorial_manager=None,
            actionSave_trigger=lambda: calls.append("save"),
            shutting_down=False,
            save_settings=lambda: calls.append("settings"),
            StopSignal=SignalRecorder(),
            http_server_thread=None,
            generation_queue=None,
            generation_service=None,
            preview_thread=None,
            preview_parent=None,
            videoPreview=None,
            timeline_sync=None,
            destroy_lock_file=lambda: calls.append("destroy_lock"),
        )
        self.app.project = types.SimpleNamespace(needs_save=lambda: True)
        self.app.logger_libopenshot = None
        event = types.SimpleNamespace(
            accept=lambda: event_calls.append("accept"),
            ignore=lambda: event_calls.append("ignore"),
        )

        with ExitStack() as stack:
            stack.enter_context(
                patch.object(
                    self.main_window_module.QMessageBox,
                    "question",
                    return_value=self.main_window_module.QMessageBox.Yes,
                )
            )
            stack.enter_context(
                patch.object(
                    self.main_window_module,
                    "track_metric_session",
                    tracker.append,
                )
            )
            stack.enter_context(
                patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None)
            )
            self.main_window_module.MainWindow.closeEvent(fake_window, event)

        self.assertEqual(event_calls, ["accept"])
        self.assertIn("save", calls)
        self.assertIn("settings", calls)
        self.assertIn("destroy_lock", calls)
        self.assertEqual(tracker, [False])
        self.assertTrue(fake_window.shutting_down)

    def test_clear_optimized_files_cancel_does_nothing(self):
        proxy_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                delete_internal_project_proxy_files=lambda: proxy_calls.append("delete"),
            ),
        )

        with patch.object(
            self.main_window_module.QMessageBox,
            "question",
            return_value=self.main_window_module.QMessageBox.No,
        ):
            self.main_window_module.MainWindow.actionClearOptimizedFiles_trigger(fake_window)

        self.assertEqual(proxy_calls, [])

    def test_clear_optimized_files_yes_deletes_project_optimized_files(self):
        proxy_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                delete_internal_project_proxy_files=lambda: proxy_calls.append("delete"),
            ),
        )

        with patch.object(
            self.main_window_module.QMessageBox,
            "question",
            return_value=self.main_window_module.QMessageBox.Yes,
        ):
            self.main_window_module.MainWindow.actionClearOptimizedFiles_trigger(fake_window)

        self.assertEqual(proxy_calls, ["delete"])

    def test_refresh_clear_menu_action_states_enables_action_only_when_internal_optimized_files_exist(self):
        enabled_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                has_internal_project_proxy_files=lambda: True,
            ),
            actionClearOptimizedFiles=types.SimpleNamespace(
                setEnabled=enabled_calls.append,
            ),
        )

        self.main_window_module.MainWindow._refresh_clear_menu_action_states(fake_window)

        self.assertEqual(enabled_calls, [True])

        enabled_calls = []
        fake_window = types.SimpleNamespace(
            proxy_service=types.SimpleNamespace(
                has_internal_project_proxy_files=lambda: False,
            ),
            actionClearOptimizedFiles=types.SimpleNamespace(
                setEnabled=enabled_calls.append,
            ),
        )

        self.main_window_module.MainWindow._refresh_clear_menu_action_states(fake_window)

        self.assertEqual(enabled_calls, [False])

    def test_close_event_no_skips_save_but_shuts_down(self):
        calls = []

        fake_window = types.SimpleNamespace(
            tutorial_manager=None,
            actionSave_trigger=lambda: calls.append("save"),
            shutting_down=False,
            save_settings=lambda: calls.append("settings"),
            StopSignal=SignalRecorder(),
            http_server_thread=None,
            generation_queue=None,
            generation_service=None,
            preview_thread=None,
            preview_parent=None,
            videoPreview=None,
            timeline_sync=None,
            destroy_lock_file=lambda: calls.append("destroy_lock"),
        )
        self.app.project = types.SimpleNamespace(needs_save=lambda: True)
        self.app.logger_libopenshot = None
        event = types.SimpleNamespace(accept=lambda: None, ignore=lambda: None)

        with ExitStack() as stack:
            stack.enter_context(
                patch.object(
                    self.main_window_module.QMessageBox,
                    "question",
                    return_value=self.main_window_module.QMessageBox.No,
                )
            )
            stack.enter_context(
                patch.object(self.main_window_module, "track_metric_session", lambda value: None)
            )
            stack.enter_context(
                patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None)
            )
            self.main_window_module.MainWindow.closeEvent(fake_window, event)

        self.assertNotIn("save", calls)
        self.assertIn("settings", calls)
        self.assertIn("destroy_lock", calls)
        self.assertTrue(fake_window.shutting_down)

    def test_open_project_success_loads_project_and_refreshes_ui(self):
        refresh_files = SignalRecorder()
        refresh_frame = SignalRecorder()
        max_size = SignalRecorder()
        recent_calls = []
        clear_temp = []
        move_calls = []

        player = types.SimpleNamespace(Seek=lambda frame: move_calls.append(("seek", frame)))
        preview_thread = types.SimpleNamespace(player=player)
        video_preview = types.SimpleNamespace(
            clearTransformState=lambda: move_calls.append(("clear_transform",)),
            size=lambda: "preview-size",
        )

        fake_window = types.SimpleNamespace(
            SpeedSignal=SignalRecorder(),
            PauseSignal=SignalRecorder(),
            videoPreview=video_preview,
            clearSelections=lambda: move_calls.append(("clear_selections",)),
            preview_thread=preview_thread,
            SetWindowTitle=lambda: recent_calls.append("title"),
            refreshFilesSignal=refresh_files,
            refreshFrameSignal=refresh_frame,
            MaxSizeChanged=max_size,
            load_recent_menu=lambda: recent_calls.append("recent"),
            clear_temporary_files=lambda: clear_temp.append(True),
            movePlayhead=lambda frame: move_calls.append(("playhead", frame)),
            actionSave_trigger=lambda: None,
            statusBar=types.SimpleNamespace(showMessage=lambda *args: None),
            remove_recent_project=lambda path: None,
        )
        load_calls = []
        history_calls = []
        self.app.project = types.SimpleNamespace(
            needs_save=lambda: False,
            load=lambda path, clear_thumbnails: load_calls.append((path, clear_thumbnails)),
        )
        self.app.updates = types.SimpleNamespace(load_history=history_calls.append)
        self.app.window = fake_window
        self.app.setOverrideCursor = lambda cursor: None
        self.app.restoreOverrideCursor = lambda: recent_calls.append("restore")

        with tempfile.TemporaryDirectory() as tmpdir:
            project_path = os.path.join(tmpdir, "existing.osp")
            with ExitStack() as stack:
                stack.enter_context(
                    patch.object(self.main_window_module.os.path, "exists", return_value=True)
                )
                stack.enter_context(
                    patch.object(self.main_window_module.QCoreApplication, "processEvents", lambda: None)
                )
                self.main_window_module.MainWindow.open_project(
                    fake_window,
                    project_path,
                    clear_thumbnails=True,
                )

            self.assertEqual(load_calls, [(project_path, True)])
            self.assertEqual(history_calls, [self.app.project])
            self.assertEqual(clear_temp, [True])
            self.assertEqual(refresh_files.calls, [()])
            self.assertEqual(refresh_frame.calls, [()])
            self.assertEqual(max_size.calls, [("preview-size",)])
            self.assertIn(("seek", 1), move_calls)
            self.assertIn(("playhead", 1), move_calls)
            self.assertIn("recent", recent_calls)
            self.assertIn("restore", recent_calls)

    def test_action_remove_clip_skips_locked_tracks(self):
        deleted = []
        removed = []
        refreshed = SignalRecorder()

        locked_clip = types.SimpleNamespace(data={"layer": 2}, delete=lambda: deleted.append("locked"))
        unlocked_clip = types.SimpleNamespace(data={"layer": 1}, delete=lambda: deleted.append("unlocked"))

        self.app.project = types.SimpleNamespace(get=lambda key: [{"number": 2, "lock": True}])
        self.app.updates = types.SimpleNamespace(transaction_id=None)
        self.app.window = types.SimpleNamespace(refreshFrameSignal=refreshed)

        fake_window = types.SimpleNamespace(
            selected_clips=["C1"],
            removeSelection=lambda item_id, item_type: removed.append((item_id, item_type)),
            emit_selection_signal=lambda: None,
            show_property_timeout=lambda: None,
        )

        with patch.object(self.main_window_module.Clip, "filter", return_value=[locked_clip, unlocked_clip]):
            self.main_window_module.MainWindow.actionRemoveClip_trigger(fake_window, refresh=True)

        self.assertEqual(deleted, ["unlocked"])
        self.assertEqual(removed, [("C1", "clip")])
        self.assertEqual(refreshed.calls, [()])
        self.assertIsNone(self.app.updates.transaction_id)

    def test_action_remove_transition_skips_locked_tracks(self):
        deleted = []
        removed = []
        refreshed = SignalRecorder()

        locked_tran = types.SimpleNamespace(data={"layer": 3}, delete=lambda: deleted.append("locked"))
        unlocked_tran = types.SimpleNamespace(data={"layer": 1}, delete=lambda: deleted.append("unlocked"))

        self.app.project = types.SimpleNamespace(get=lambda key: [{"number": 3, "lock": True}])
        self.app.updates = types.SimpleNamespace(transaction_id=None)
        self.app.window = types.SimpleNamespace(refreshFrameSignal=refreshed)

        fake_window = types.SimpleNamespace(
            selected_transitions=["T1"],
            removeSelection=lambda item_id, item_type: removed.append((item_id, item_type)),
            emit_selection_signal=lambda: None,
            show_property_timeout=lambda: None,
            refreshFrameSignal=refreshed,
        )

        with patch.object(self.main_window_module.Transition, "filter", return_value=[locked_tran, unlocked_tran]):
            self.main_window_module.MainWindow.actionRemoveTransition_trigger(fake_window, refresh=True)

        self.assertEqual(deleted, ["unlocked"])
        self.assertEqual(removed, [("T1", "transition")])
        self.assertEqual(refreshed.calls, [()])
        self.assertIsNone(self.app.updates.transaction_id)

    def test_ripple_delete_gap_shifts_only_later_items_on_same_layer(self):
        saved = []
        clips = [
            types.SimpleNamespace(data={"position": 4.0}, save=lambda: saved.append("clip-before")),
            types.SimpleNamespace(data={"position": 9.0}, save=lambda: saved.append("clip-after")),
        ]
        transitions = [
            types.SimpleNamespace(data={"position": 8.0}, save=lambda: saved.append("tran-before")),
            types.SimpleNamespace(data={"position": 12.0}, save=lambda: saved.append("tran-after")),
        ]

        with ExitStack() as stack:
            stack.enter_context(patch.object(self.main_window_module.Clip, "filter", return_value=clips))
            stack.enter_context(
                patch.object(self.main_window_module.Transition, "filter", return_value=transitions)
            )
            self.main_window_module.MainWindow.ripple_delete_gap(
                types.SimpleNamespace(),
                ripple_start=8.0,
                layer=1,
                total_gap=2.5,
            )

        self.assertEqual(clips[0].data["position"], 4.0)
        self.assertEqual(clips[1].data["position"], 6.5)
        self.assertEqual(transitions[0].data["position"], 8.0)
        self.assertEqual(transitions[1].data["position"], 9.5)
        self.assertEqual(saved, ["clip-after", "tran-after"])



"""
 @file
 @brief This file contains the effects model, used by the main window
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os

from PyQt5.QtCore import (
    QObject, QMimeData, Qt, QSize, pyqtSignal,
    QSortFilterProxyModel, QPersistentModelIndex, QItemSelectionModel, QItemSelection, QModelIndex,
)
from PyQt5.QtGui import QIcon, QPixmap, QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QMessageBox

import openshot  # Python module for libopenshot (required video editing module installed separately)

from classes import info
from classes.logger import log
from classes.app import get_app

import json


class SingleColumnProxyModel(QSortFilterProxyModel):
    """Proxy that exposes only the first column for ListView accessibility"""

    def columnCount(self, parent=QModelIndex()):
        return 1

    def data(self, index, role=Qt.DisplayRole):
        """Get text data from the underlying source model (bypassing filter proxy)"""
        if index.column() == 0 and role in (Qt.DisplayRole, Qt.AccessibleTextRole):
            source_index = self.mapToSource(index)
            filter_proxy = self.sourceModel()
            if filter_proxy:
                root_index = filter_proxy.mapToSource(source_index)
                root_model = filter_proxy.sourceModel()
                if root_model:
                    return root_model.data(root_index, Qt.DisplayRole)
        return super().data(index, role)


class EffectsProxyModel(QSortFilterProxyModel):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def data(self, index, role=Qt.DisplayRole):
        """Hide text in column 0 for TreeView - name is shown in column 1"""
        if index.column() == 0 and role in (Qt.DisplayRole, Qt.AccessibleTextRole):
            return ""
        return super().data(index, role)

    def filterAcceptsRow(self, sourceRow, sourceParent):
        """Filter for common transitions and text filter"""

        if not get_app().window.actionEffectsShowAll.isChecked():
            # Fetch the effect values
            effect_name = self.sourceModel().data(self.sourceModel().index(sourceRow, 1, sourceParent))
            effect_desc = self.sourceModel().data(self.sourceModel().index(sourceRow, 2, sourceParent))
            effect_type = self.sourceModel().data(self.sourceModel().index(sourceRow, 3, sourceParent))

            # Return, if regExp match in displayed format.
            if get_app().window.actionEffectsShowVideo.isChecked():
                return effect_type == "Video" and \
                    (self.filterRegExp().indexIn(effect_name) >= 0 or
                     self.filterRegExp().indexIn(effect_desc) >= 0)
            else:
                return effect_type == "Audio" and \
                    (self.filterRegExp().indexIn(effect_name) >= 0 or
                     self.filterRegExp().indexIn(effect_desc) >= 0)

        # Continue running built-in parent filter logic
        return super(EffectsProxyModel, self).filterAcceptsRow(sourceRow, sourceParent)

    def mimeData(self, indexes):
        # Create MimeData for drag operation
        data = QMimeData()

        # Get list of class names for requested effect indexes
        items = []
        for proxy_index in indexes:
            source_index = self.mapToSource(proxy_index)
            items.append(source_index.sibling(source_index.row(), 4).data())
        data.setText(json.dumps(items))
        data.setHtml("effect")

        # Return Mimedata
        return data


class EffectsModel(QObject):
    ModelRefreshed = pyqtSignal()

    def update_model(self, clear=True):
        log.info("updating effects model.")
        app = get_app()

        # Get window to check filters
        win = app.window
        _ = app._tr

        # Clear all items
        if clear:
            self.model_names = {}
            self.model.clear()

        # Add Headers
        self.model.setHorizontalHeaderLabels([_("Thumb"), _("Name"), _("Description")])

        # Get the folder path of effects
        effects_dir = os.path.join(info.PATH, "effects")
        icons_dir = os.path.join(effects_dir, "icons")

        # Get a JSON list of all supported effects in libopenshot
        raw_effects_list = json.loads(openshot.EffectInfo.Json())

        # Loop through each effect
        for effect_info in raw_effects_list:
            # Get basic properties about each effect
            effect_name = effect_info["class_name"]
            title = effect_info["name"]
            description = effect_info["description"]
            # Remove any spaces from icon name
            icon_name = "%s.png" % effect_name.lower().replace(' ', '')
            icon_path = os.path.join(icons_dir, icon_name)

            # Determine the category of effect (audio, video, both)
            category = None
            if effect_info["has_video"] and effect_info["has_audio"]:
                category = "Audio & Video"
            elif not effect_info["has_video"] and effect_info["has_audio"]:
                category = "Audio"
            elif effect_info["has_video"] and not effect_info["has_audio"]:
                category = "Video"

            # Check for thumbnail path (in build-in cache)
            thumb_path = os.path.join(info.IMAGES_PATH, "cache", icon_name)

            # Check built-in cache (if not found)
            if not os.path.exists(thumb_path):
                # Check user folder cache
                thumb_path = os.path.join(info.CACHE_PATH, icon_name)

            # Generate thumbnail (if needed)
            if not os.path.exists(thumb_path):

                try:
                    # Reload this reader
                    log.info('Generating thumbnail for %s (%s)' % (thumb_path, icon_path))
                    clip = openshot.Clip(icon_path)
                    reader = clip.Reader()

                    # Open reader
                    reader.Open()

                    # Save thumbnail
                    reader.GetFrame(0).Thumbnail(
                        thumb_path, 98, 64,
                        os.path.join(info.IMAGES_PATH, "mask.png"),
                        "", "#000", True, "png", 85
                    )
                    reader.Close()

                except Exception:
                    # Handle exception
                    log.warning("{} is not a valid image file.".format(icon_path))

            row = []

            # Append thumbnail
            col = QStandardItem()

            # Load icon (using display DPI)
            icon = QIcon()
            icon.addFile(thumb_path)

            col.setIcon(icon)
            col.setText(self.app._tr(title))
            col.setToolTip(self.app._tr(title))
            col.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable | Qt.ItemIsDragEnabled)
            col.setAccessibleText(self.app._tr(title))
            row.append(col)

            # Append Name
            col = QStandardItem("Name")
            col.setData(self.app._tr(title), Qt.DisplayRole)
            col.setText(self.app._tr(title))
            col.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable | Qt.ItemIsDragEnabled)
            col.setAccessibleText(self.app._tr(title))
            row.append(col)

            # Append Description
            col = QStandardItem("Description")
            col.setData(self.app._tr(description), Qt.DisplayRole)
            col.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable | Qt.ItemIsDragEnabled)
            row.append(col)

            # Append Category
            col = QStandardItem("Category")
            col.setData(category, Qt.DisplayRole)
            col.setText(category)
            col.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable | Qt.ItemIsDragEnabled)
            row.append(col)

            # Append Path
            col = QStandardItem("Effect")
            col.setData(effect_name, Qt.DisplayRole)
            col.setText(effect_name)
            col.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable | Qt.ItemIsDragEnabled)
            row.append(col)

            # Append ROW to MODEL (if does not already exist in model)
            if effect_name not in self.model_names:
                self.model.appendRow(row)
                self.model_names[effect_name] = QPersistentModelIndex(row[1].index())

        # Emit signal when model is updated
        self.ModelRefreshed.emit()

    def _sync_tree_to_list_selection(self, selected, deselected):
        """Sync selection from TreeView (proxy_model) to ListView (list_proxy_model)"""
        if self._syncing_selection:
            return
        self._syncing_selection = True
        try:
            list_selection = QItemSelection()
            for index in self.selection_model.selectedRows(0):
                list_index = self.list_proxy_model.mapFromSource(index)
                if list_index.isValid():
                    list_selection.select(list_index, list_index)
            self.list_selection_model.select(
                list_selection,
                QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
            )
        finally:
            self._syncing_selection = False

    def _sync_list_to_tree_selection(self, selected, deselected):
        """Sync selection from ListView (list_proxy_model) to TreeView (proxy_model)"""
        if self._syncing_selection:
            return
        self._syncing_selection = True
        try:
            tree_selection = QItemSelection()
            for index in self.list_selection_model.selectedRows(0):
                tree_index = self.list_proxy_model.mapToSource(index)
                if tree_index.isValid():
                    tree_selection.select(tree_index, tree_index)
            self.selection_model.select(
                tree_selection,
                QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
            )
        finally:
            self._syncing_selection = False

    def __init__(self, *args):
        # Init QObject superclass
        super().__init__(*args)

        # Create standard model
        self.app = get_app()
        self.model = QStandardItemModel()
        self.model.setColumnCount(5)
        self.model_names = {}

        # Create proxy model (for sorting and filtering) - used by TreeView
        self.proxy_model = EffectsProxyModel()
        self.proxy_model.setDynamicSortFilter(True)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseSensitive)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setSortLocaleAware(True)
        self.proxy_model.setFilterKeyColumn(-1)

        # Create single-column proxy for ListView (wraps proxy_model for accessibility)
        self.list_proxy_model = SingleColumnProxyModel()
        self.list_proxy_model.setSourceModel(self.proxy_model)

        # Create selection models for each view
        self.selection_model = QItemSelectionModel(self.proxy_model)
        self.list_selection_model = QItemSelectionModel(self.list_proxy_model)

        # Sync selections between the two selection models
        self._syncing_selection = False
        self.selection_model.selectionChanged.connect(self._sync_tree_to_list_selection)
        self.list_selection_model.selectionChanged.connect(self._sync_list_to_tree_selection)

        # Attempt to load model testing interface, if requested
        # (will only succeed with Qt 5.11+)
        if info.MODEL_TEST:
            try:
                # Create model tester objects
                from PyQt5.QtTest import QAbstractItemModelTester
                self.model_tests = []
                for m in [self.proxy_model, self.model]:
                    self.model_tests.append(
                        QAbstractItemModelTester(
                            m, QAbstractItemModelTester.FailureReportingMode.Warning)
                    )
                log.info("Enabled {} model tests for effects data".format(len(self.model_tests)))
            except ImportError:
                pass



"""
 @file
 @brief Clip timing and frame utilities shared by the UI.

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import logging
import json
from fractions import Fraction
from typing import Any, Mapping, Optional, Tuple

import openshot

from classes.app import get_app

logger = logging.getLogger(__name__)


def apply_file_caption_to_clip(clip_data: Any, file_obj: Any, *, dedupe: bool = True) -> bool:
    """Attach a Caption effect to clip_data when file metadata includes caption text."""
    if not isinstance(clip_data, Mapping):
        return False
    file_data = getattr(file_obj, "data", None)
    if not isinstance(file_data, Mapping):
        return False
    caption_text = str(file_data.get("caption", "") or "").strip()
    if not caption_text:
        return False

    effects = clip_data.get("effects")
    if not isinstance(effects, list):
        effects = list(effects) if effects else []
        clip_data["effects"] = effects

    if dedupe:
        for effect in effects:
            if not isinstance(effect, Mapping):
                continue
            if str(effect.get("class_name", "")).lower() == "caption":
                existing_text = str(effect.get("caption_text", "") or "").strip()
                if existing_text == caption_text:
                    return False

    caption_effect = openshot.EffectInfo().CreateEffect("Caption")
    caption_effect.Id(get_app().project.generate_id())
    caption_json = json.loads(caption_effect.Json())
    caption_json["caption_text"] = caption_text
    effects.append(caption_json)
    return True


def _as_mapping(candidate: Any) -> Mapping[str, Any]:
    """Return dict-style metadata for clips, readers, or similar."""
    if isinstance(candidate, Mapping):
        return candidate
    data = getattr(candidate, "data", None)
    if isinstance(data, Mapping):
        return data
    return {}


def _rounded_int(value: Any) -> Optional[int]:
    """Round a numeric value to int when possible."""
    try:
        return int(round(float(value)))
    except (TypeError, ValueError):
        return None


def _to_float(value: Any) -> Optional[float]:
    """Convert a value to float when possible."""
    try:
        return float(value)
    except (TypeError, ValueError):
        return None


def _to_positive_float(value: Any) -> Optional[float]:
    """Convert a value to a positive float."""
    number = _to_float(value)
    if number is None or number <= 0:
        return None
    return number


def _to_positive_int(value: Any) -> Optional[int]:
    """Return a rounded positive int."""
    number = _rounded_int(value)
    if number is None or number <= 0:
        return None
    return number


def _fps_fraction(fps_value: Any) -> Optional[Fraction]:
    """Convert an FPS value to a Fraction."""
    if fps_value is None:
        return None
    if isinstance(fps_value, Fraction):
        return fps_value
    if isinstance(fps_value, (int, float)):
        if fps_value > 0:
            return Fraction(fps_value).limit_denominator(1_000_000)
        return None
    if isinstance(fps_value, Mapping):
        fps_num = fps_value.get("num")
        fps_den = fps_value.get("den")
    else:
        fps_num = getattr(fps_value, "num", None)
        fps_den = getattr(fps_value, "den", None)
    try:
        fps_num = int(fps_num)
        fps_den = int(fps_den)
    except (TypeError, ValueError):
        fps_num = fps_den = None
    if fps_num and fps_den:
        try:
            return Fraction(fps_num, fps_den)
        except ZeroDivisionError:
            return None
    to_float = getattr(fps_value, "ToFloat", None)
    if callable(to_float):
        fps_float = to_float()
        if fps_float and fps_float > 0:
            return Fraction(fps_float).limit_denominator(1_000_000)
    return None


def project_fps_fraction() -> Fraction:
    """Return the current project FPS as a Fraction."""
    app = get_app()
    project = getattr(app, "project", None) if app else None
    fps_meta = None
    if hasattr(project, "get"):
        try:
            fps_meta = project.get("fps")
        except TypeError:
            fps_meta = None
    elif isinstance(project, Mapping):
        fps_meta = project.get("fps")
    fps = _fps_fraction(fps_meta)
    return fps or Fraction(30, 1)


def _project_fps_float() -> float:
    """Return the project FPS as a float."""
    try:
        return float(project_fps_fraction())
    except (TypeError, ValueError):
        return 30.0


def video_length_to_project_frames(
    media: Any = None,
    *,
    video_length: Any = None,
    fps: Any = None,
    duration: Any = None,
    project_fps: Any = None,
) -> Optional[int]:
    """Return project frames needed to play supplied media."""
    metadata = _as_mapping(media)
    if video_length is None:
        video_length = metadata.get("video_length")
    frames = _to_positive_int(video_length)

    if fps is None:
        fps = metadata.get("fps")
    source_fps = _fps_fraction(fps)

    if duration is None:
        duration = metadata.get("duration")
    duration_value = _to_positive_float(duration)

    if frames is None and duration_value is not None and source_fps:
        frames = int(round(duration_value * float(source_fps)))

    if frames is None or frames <= 0:
        return None

    project_fraction = _fps_fraction(project_fps) or project_fps_fraction()

    if not source_fps or not project_fraction:
        return max(frames, 1)

    scaled = Fraction(frames) * project_fraction / source_fps
    try:
        scaled_value = int(round(float(scaled)))
    except (TypeError, ValueError):
        return max(frames, 1)
    return max(scaled_value, 1)


def _clip_reader(clip_data: Any, existing_clip: Any) -> Any:
    """Return reader metadata sourced from clip data or existing clip."""
    reader = _as_mapping(clip_data).get("reader")
    if reader:
        return reader
    return _as_mapping(existing_clip).get("reader")


def _inherit_timing(clip_data: Mapping[str, Any], existing_clip: Any) -> None:
    """Fill missing timing fields from an existing clip."""
    existing = _as_mapping(existing_clip)
    if not existing:
        return
    for key in ("start", "end", "duration"):
        if key not in clip_data or clip_data.get(key) is None:
            if key in existing:
                clip_data[key] = existing.get(key)


def _clip_id(clip_data: Any, existing_clip: Any) -> Optional[str]:
    """Return the clip id from data or an existing clip."""
    clip_meta = _as_mapping(clip_data)
    clip_id = clip_meta.get("id")
    if clip_id:
        return clip_id
    existing_meta = _as_mapping(existing_clip)
    clip_id = existing_meta.get("id")
    return clip_id if clip_id else None


def _timeline_clip(clip_data: Any, existing_clip: Any) -> Any:
    """Return the live timeline clip instance when available."""
    clip_id = _clip_id(clip_data, existing_clip)
    if not clip_id:
        return None
    try:
        app = get_app()
        window = getattr(app, "window", None)
        timeline_sync = getattr(window, "timeline_sync", None)
        timeline = getattr(timeline_sync, "timeline", None)
        return timeline.GetClip(clip_id) if timeline else None
    except Exception as exc:  # pragma: no cover - defensive logging
        logger.debug("Unable to locate clip %s on timeline: %s", clip_id, exc, exc_info=True)
        return None


def _time_points(clip_data: Any) -> Optional[list]:
    """Return time curve points from clip metadata."""
    time_data = _as_mapping(clip_data).get("time")
    points = time_data.get("Points") if isinstance(time_data, Mapping) else None
    return points if isinstance(points, list) else None


def _time_curve_length_frames(clip_data: Any, existing_clip: Any) -> Optional[int]:
    """Return the max frame referenced by the time curve."""
    points = _time_points(clip_data)
    if points:
        max_x = 0
        for point in points:
            co = point.get("co") if isinstance(point, Mapping) else None
            if not isinstance(co, Mapping):
                continue
            x_val = _rounded_int(co.get("X"))
            if x_val is not None:
                max_x = max(max_x, x_val)
        if max_x:
            return max_x

    clip_obj = _timeline_clip(clip_data, existing_clip)
    if clip_obj and getattr(clip_obj, "time", None):
        try:
            length = clip_obj.time.GetLength()
        except Exception as exc:  # pragma: no cover - defensive logging
            logger.debug("Unable to query clip time length: %s", exc, exc_info=True)
        else:
            frame_count = _rounded_int(length)
            if frame_count:
                return max(frame_count, 1)

    return None


def _clamp_time_points(points: Any, max_frames: Optional[int]) -> None:
    """Clamp keyframe points to positive axes and available frames."""
    if not isinstance(points, list) or not max_frames:
        return

    limit = int(max_frames)
    for point in points:
        co = point.get("co") if isinstance(point, Mapping) else None
        if not isinstance(co, Mapping):
            continue
        x_val = _rounded_int(co.get("X"))
        if x_val is not None:
            co["X"] = x_val
        y_val = _rounded_int(co.get("Y"))
        if y_val is not None:
            if y_val < 1:
                co["Y"] = 1
            elif y_val > limit:
                co["Y"] = limit
            else:
                co["Y"] = y_val


def _clamp_basic_timing(clip_data: Mapping[str, Any], max_duration: Optional[float]) -> None:
    """Clamp start/end/duration for clips without remapping."""
    start = _to_float(clip_data.get("start"))
    if start is None or start < 0.0:
        start = 0.0
    if max_duration is not None and start > max_duration:
        start = max_duration

    end = _to_float(clip_data.get("end"))
    duration = _to_float(clip_data.get("duration"))

    if end is None:
        if duration is not None:
            end = start + max(duration, 0.0)
        elif max_duration is not None:
            end = max_duration
        else:
            end = start

    if max_duration is not None and end > max_duration:
        end = max_duration
    if end < start:
        start = end

    clip_data["start"] = start
    clip_data["end"] = end
    clip_data["duration"] = max(0.0, end - start)


def _clamp_curve_start(clip_data: Mapping[str, Any], max_duration: Optional[float]) -> None:
    """Clamp the start value for curves to the available range."""
    start = _to_float(clip_data.get("start"))
    if start is None or start < 0.0:
        start = 0.0
    if max_duration is not None and start > max_duration:
        start = max_duration
    clip_data["start"] = start


def _is_single_image(source: Any) -> bool:
    """Return True when metadata flags the media as a single image."""
    metadata = _as_mapping(source)
    if metadata.get("has_single_image"):
        if is_audio_media(metadata):
            return False
        return True
    media_type = metadata.get("media_type")
    return isinstance(media_type, str) and media_type.lower() == "image"


def is_audio_media(source: Any) -> bool:
    """Return True when metadata likely represents an audio asset."""
    metadata = _as_mapping(source)

    media_type = str(metadata.get("media_type") or "").strip().lower()
    if media_type == "audio":
        return True

    source_path = str(
        metadata.get("path")
        or metadata.get("file_path")
        or ""
    ).strip().lower()
    audio_exts = (".mp3", ".m4a", ".aac", ".ogg", ".opus", ".flac", ".wav", ".wma")
    if source_path.endswith(audio_exts):
        return True

    has_audio = metadata.get("has_audio")
    has_video = metadata.get("has_video")
    if has_audio is True and has_video is False:
        return True

    return False


def is_single_image_media(source: Any) -> bool:
    """Public helper to identify still-image media metadata."""
    return _is_single_image(source)


def _clip_has_single_image(reader: Any, clip_data: Any, existing_clip: Any) -> bool:
    """Return True if any metadata indicates a single-image clip."""
    if _is_single_image(reader):
        return True
    clip_meta = _as_mapping(clip_data)
    if _is_single_image(clip_meta):
        return True
    if _is_single_image(clip_meta.get("reader")):
        return True
    existing_meta = _as_mapping(existing_clip)
    if _is_single_image(existing_meta):
        return True
    return _is_single_image(existing_meta.get("reader"))


def _normalize_single_image(clip_data: Mapping[str, Any]) -> None:
    """Set consistent timing for single-image clips."""
    start = _to_float(clip_data.get("start"))
    if start is None or start < 0.0:
        start = 0.0

    end = _to_float(clip_data.get("end"))
    duration = _to_float(clip_data.get("duration"))

    if duration is None and end is not None:
        duration = end - start
    if end is None and duration is not None:
        end = start + duration
    if end is None:
        end = start
    if duration is None:
        duration = end - start

    if end < start:
        end = start
    duration = max(0.0, end - start)

    clip_data["start"] = start
    clip_data["end"] = end
    clip_data["duration"] = duration


def _reader_bounds(reader: Any) -> Tuple[Optional[float], Optional[int]]:
    """Return reader duration and total frames in project units."""
    metadata = _as_mapping(reader)
    duration = _to_float(metadata.get("duration"))
    if duration is not None and duration < 0.0:
        duration = None

    project_fps = project_fps_fraction()
    project_fps_float = float(project_fps) if project_fps else None

    frames = video_length_to_project_frames(metadata, project_fps=project_fps)

    if duration is None and frames and project_fps_float:
        duration = frames / project_fps_float
    if frames is None and duration is not None and project_fps_float:
        frames = _to_positive_int(duration * project_fps_float)

    return duration, frames


def clamp_timing_to_media(clip_data: Mapping[str, Any], existing_clip: Any = None) -> Mapping[str, Any]:
    """Clamp clip timing to the available source media or time-curve bounds."""
    reader = _clip_reader(clip_data, existing_clip)

    _inherit_timing(clip_data, existing_clip)

    if _clip_has_single_image(reader, clip_data, existing_clip):
        _normalize_single_image(clip_data)
        return clip_data

    points = _time_points(clip_data)
    multi_point_time = isinstance(points, list) and len(points) > 1

    reader_duration, reader_frames = _reader_bounds(reader)

    if multi_point_time:
        max_frames = _time_curve_length_frames(clip_data, existing_clip) or reader_frames
        if reader_frames:
            _clamp_time_points(points, reader_frames)

        proj_fps_f = _project_fps_float()
        max_duration = None
        if max_frames and proj_fps_f:
            max_duration = max_frames / proj_fps_f
        elif reader_duration is not None:
            max_duration = reader_duration

        _clamp_curve_start(clip_data, max_duration)
        return clip_data

    _clamp_basic_timing(clip_data, reader_duration)
    if reader_frames:
        _clamp_time_points(points, reader_frames)

    return clip_data


def clip_time_bounds(clip_data: Any, existing_clip: Any = None) -> Tuple[float, int]:
    """Return the max duration (seconds) and frame count allowed for the clip."""
    reader = _clip_reader(clip_data, existing_clip) or {}

    if _clip_has_single_image(reader, clip_data, existing_clip):
        clip_meta = _as_mapping(clip_data)
        duration = _to_float(clip_meta.get("duration"))
        if duration is None:
            duration = _to_float(_as_mapping(reader).get("duration"))
        duration = duration if duration is not None else 0.0
        proj_fps_f = _project_fps_float()
        if duration and proj_fps_f:
            max_frames = max(1, int(round(duration * proj_fps_f)))
        else:
            max_frames = 1
        return duration, max_frames

    points = _time_points(clip_data)
    multi_point_time = isinstance(points, list) and len(points) > 1

    proj_fps_f = _project_fps_float()
    reader_duration, reader_frames = _reader_bounds(reader)

    if multi_point_time:
        max_frames = _time_curve_length_frames(clip_data, existing_clip) or reader_frames or 1
        if proj_fps_f:
            max_duration = max_frames / proj_fps_f
        elif reader_duration is not None:
            max_duration = reader_duration
        else:
            max_duration = 0.0
        return max_duration, max_frames

    max_frames = reader_frames or 1
    if reader_duration is not None:
        max_duration = reader_duration
    elif proj_fps_f and max_frames:
        max_duration = max_frames / proj_fps_f
    else:
        max_duration = 0.0

    return max_duration, max_frames



"""
 @file
 @brief This file contains the project file model, used by the project tree
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import json
import re
import glob
import functools
import uuid

from PyQt5.QtCore import (
    QMimeData, Qt, pyqtSignal, QEventLoop, QObject,
    QSortFilterProxyModel, QItemSelectionModel, QItemSelection, QPersistentModelIndex, QModelIndex
)
from PyQt5.QtGui import (
    QIcon, QPixmap, QStandardItem, QStandardItemModel
)
from PyQt5.QtWidgets import QAbstractItemView
from classes import updates
from classes import info
from classes.image_types import get_media_type
from classes.query import File
from classes.logger import log
from classes.app import get_app
from classes.thumbnail import GetThumbPath
from classes.qt_types import model_index_sibling_at_column

import openshot


IMPORT_READER_MAX_SIZE = 128


def inspect_media(path, max_width=0, max_height=0):
    """Inspect media using the shared libopenshot reader-selection logic."""
    reader = openshot.Clip.CreateReader(path, False)
    if not reader:
        raise RuntimeError(f"No reader available for path: {path}")
    if max_width > 0 and max_height > 0 and hasattr(reader, "SetMaxDecodeSize"):
        reader.SetMaxDecodeSize(int(max_width), int(max_height))
    reader.Open()
    try:
        return json.loads(reader.Json()), float(reader.info.duration or 0.0)
    finally:
        reader.Close()


class SingleColumnProxyModel(QSortFilterProxyModel):
    """Proxy that exposes only the first column for ListView accessibility"""

    def columnCount(self, parent=QModelIndex()):
        return 1

    def data(self, index, role=Qt.DisplayRole):
        """Get text data from the underlying source model (bypassing filter proxy)"""
        if index.column() == 0 and role in (Qt.DisplayRole, Qt.AccessibleTextRole):
            # Get the actual text from the root source model (QStandardItemModel)
            # by traversing through the proxy chain
            source_index = self.mapToSource(index)
            filter_proxy = self.sourceModel()
            if filter_proxy:
                root_index = filter_proxy.mapToSource(source_index)
                root_model = filter_proxy.sourceModel()
                if root_model:
                    return root_model.data(root_index, Qt.DisplayRole)
        return super().data(index, role)


class FileFilterProxyModel(QSortFilterProxyModel):
    """Proxy class used for sorting and filtering model data"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def data(self, index, role=Qt.DisplayRole):
        """Hide text in column 0 for TreeView - name is shown in column 1"""
        if index.column() == 0 and role in (Qt.DisplayRole, Qt.AccessibleTextRole):
            return ""
        return super().data(index, role)

    def filterAcceptsRow(self, sourceRow, sourceParent):
        """Filter for text"""
        if get_app().window.actionFilesShowVideo.isChecked() \
                or get_app().window.actionFilesShowAudio.isChecked() \
                or get_app().window.actionFilesShowImage.isChecked() \
                or get_app().window.filesFilter.text():
            # Fetch the file name
            index = self.sourceModel().index(sourceRow, 0, sourceParent)
            file_name = self.sourceModel().data(index)  # file name (i.e. MyVideo.mp4)

            # Fetch the media_type
            index = self.sourceModel().index(sourceRow, 3, sourceParent)
            media_type = self.sourceModel().data(index)  # media type (i.e. video, image, audio)

            index = self.sourceModel().index(sourceRow, 2, sourceParent)
            tags = self.sourceModel().data(index)  # tags (i.e. intro, custom, etc...)

            if any([
                get_app().window.actionFilesShowVideo.isChecked() and media_type != "video",
                get_app().window.actionFilesShowAudio.isChecked() and media_type != "audio",
                get_app().window.actionFilesShowImage.isChecked() and media_type != "image",
            ]):
                return False

            # Match against regex pattern
            return self.filterRegExp().indexIn(file_name) >= 0 or self.filterRegExp().indexIn(tags) >= 0

        # Continue running built-in parent filter logic
        return super().filterAcceptsRow(sourceRow, sourceParent)

    def mimeData(self, indexes):
        # Create MimeData for drag operation
        data = QMimeData()

        # Get list of all selected file ids
        ids = self.parent.selected_file_ids()
        data.setText(json.dumps(ids))
        data.setHtml("clip")

        # Return Mimedata
        return data

    def get_file_index(self, file_id):
        # Find the index in the proxy model based on the file ID
        if file_id in self.parent.model_ids:
            return self.mapFromSource(QModelIndex(self.parent.model_ids[file_id]))
        return QModelIndex()

    def __init__(self, **kwargs):
        if "parent" in kwargs:
            self.parent = kwargs["parent"]
            kwargs.pop("parent")

        # Call base class implementation
        super().__init__(**kwargs)


class FilesModel(QObject, updates.UpdateInterface):
    ModelRefreshed = pyqtSignal()
    PLACEHOLDER_PREFIX = "__genjob__:"
    PROJECT_FILE_THUMB_ATTEMPTS = 3

    @staticmethod
    def _icon_from_thumbnail_source(thumb_source):
        """Create an icon from freshly loaded thumbnail bytes when possible."""
        thumb_source = str(thumb_source or "")
        if thumb_source:
            pixmap = QPixmap()
            if pixmap.load(thumb_source) and not pixmap.isNull():
                return QIcon(pixmap)
        return QIcon(thumb_source)

    def _thumbnail_source_for_file(self, file, clear_cache=False):
        """Return the thumbnail/artwork source path and display name for a file."""
        path, filename = os.path.split(file.data["path"])
        name = file.data.get("name", filename)
        media_type = file.data.get("media_type")

        if media_type in ["video", "image"]:
            thumbnail_frame = 1
            if 'start' in file.data:
                fps = file.data["fps"]
                fps_float = float(fps["num"]) / float(fps["den"])
                thumbnail_frame = round(float(file.data['start']) * fps_float) + 1
            thumb_source = GetThumbPath(
                file.id,
                thumbnail_frame,
                clear_cache=clear_cache,
                attempts=self.PROJECT_FILE_THUMB_ATTEMPTS,
            )
        else:
            thumb_source = os.path.join(info.PATH, "images", "AudioThumbnail.svg")

        return thumb_source, name, media_type

    def _project_file_icon_for_file(self, file):
        thumb_source, name, media_type = self._thumbnail_source_for_file(file)
        return self._icon_from_thumbnail_source(thumb_source), name, media_type

    def _tooltip_for_file(self, file, name):
        tooltip = str(name or "")
        app = get_app()
        window = getattr(app, "window", None) if app else None
        proxy_service = getattr(window, "proxy_service", None) if window else None
        if not proxy_service or not file:
            return tooltip

        state = proxy_service.get_proxy_state(file)
        if state == "ready":
            return "{} {}".format(tooltip, app._tr("(Optimized)"))
        if state == "missing":
            return "{} {}".format(tooltip, app._tr("(Optimized)"))
        return tooltip

    # This method is invoked by the UpdateManager each time a change happens (i.e UpdateInterface)
    def changed(self, action):

        # Something was changed in the 'files' list
        if action and ((len(action.key) >= 1 and action.key[0].lower() == "files") or action.type == "load"):
            # Refresh project files model
            if action.type == "insert":
                # Don't clear the existing items if only inserting new things
                self.update_model(clear=False)
            elif action.type == "delete" and action.key[0].lower() == "files" and len(action.key) == 2:
                # Delete a top-level file row only when the file object itself was deleted.
                self.update_model(clear=False, delete_file_id=action.key[1].get('id', ''))
            elif action.type in ("update", "delete") and action.key[0].lower() == "files":
                # Update a single file (if found)
                self.update_model(clear=False, update_file_id=action.key[1].get('id', ''))
            else:
                # Clear existing items. For full project loads, batch updates for faster UI rebuild.
                self.update_model(clear=True, progressive_ui=False)

    def update_model(self, clear=True, delete_file_id=None, update_file_id=None, progressive_ui=True):
        log.debug("updating files model.")
        app = get_app()

        self.ignore_updates = True

        # Translations
        _ = app._tr

        # Delete a file (if delete_file_id passed in)
        if delete_file_id in self.model_ids:
            # Use the persistent index we stored to find the row
            id_index = self.model_ids[delete_file_id]

            # sanity check
            if not id_index.isValid() or delete_file_id != id_index.data():
                log.warning("Couldn't remove {} from model!".format(delete_file_id))
                return
            # Delete row from model
            row_num = id_index.row()
            self.model.removeRows(row_num, 1, id_index.parent())
            self.model.submit()
            self.model_ids.pop(delete_file_id)

        # Update a file (if update_file_id passed in)
        if update_file_id in self.model_ids:
            # Use the persistent index we stored to find the row
            id_index = self.model_ids[update_file_id]

            # sanity check
            if not id_index.isValid() or update_file_id != id_index.data():
                log.warning("Couldn't update {} in model!".format(update_file_id))
                return

            # lookup File object
            f = File.get(id=update_file_id)
            if f:
                # Update "tags" in model (if different)
                row_num = id_index.row()
                if f.data.get("tags") != self.model.item(row_num, 2).text():
                    self.model.item(row_num, 2).setText(f.data.get("tags"))
                path, filename = os.path.split(f.data["path"])
                name = f.data.get("name", filename)
                self.model.item(row_num, 0).setToolTip(self._tooltip_for_file(f, name))

        # Clear all items
        if clear:
            self.model_ids = {}
            self.model.clear()

        # Add Headers (all 6 columns - last 3 are hidden but must exist for proper layout)
        self.model.setHorizontalHeaderLabels([
            _("Thumb"), _("Name"), _("Tags"),
            "media_type", "path", "id"
        ])

        # Get list of files in project
        files = File.filter()  # get all files

        # add item for each file
        row_added_count = 0
        for file in files:
            id = file.data["id"]
            if id in self.model_ids and self.model_ids[id].isValid():
                # Ignore files that already exist in model
                continue

            path, filename = os.path.split(file.data["path"])
            tags = file.data.get("tags", "")
            thumb_icon, name, media_type = self._project_file_icon_for_file(file)

            row = []
            flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDragEnabled | Qt. ItemNeverHasChildren

            # Append thumbnail
            col = QStandardItem(thumb_icon, name)
            col.setToolTip(self._tooltip_for_file(file, name))
            col.setFlags(flags)
            col.setAccessibleText(name)
            row.append(col)

            # Append Filename
            col = QStandardItem(name)
            col.setFlags(flags | Qt.ItemIsEditable)
            col.setAccessibleText(name)
            row.append(col)

            # Append Tags
            col = QStandardItem(tags)
            col.setFlags(flags | Qt.ItemIsEditable)
            row.append(col)

            # Append Media Type
            col = QStandardItem(media_type)
            col.setFlags(flags)
            row.append(col)

            # Append Path
            col = QStandardItem(path)
            col.setFlags(flags)
            row.append(col)

            # Append ID
            col = QStandardItem(id)
            col.setFlags(flags | Qt.ItemIsUserCheckable)
            row.append(col)

            # Append ROW to MODEL (if does not already exist in model)
            if id not in self.model_ids:
                self.model.appendRow(row)
                # Link the file ID hash to that column of the table row by persistent index
                self.model_ids[id] = QPersistentModelIndex(row[5].index())

                row_added_count += 1
                if progressive_ui and row_added_count % 25 == 0:
                    # Update every X items
                    get_app().processEvents(QEventLoop.ExcludeUserInputEvents)

            # Refresh view/filtering incrementally during interactive updates (i.e. imports)
            if progressive_ui:
                get_app().window.resize_contents()

        self.ignore_updates = False

        # Single refresh after bulk updates (i.e. opening a project)
        if not progressive_ui:
            get_app().window.resize_contents()

        # Emit signal when model is updated
        self.ModelRefreshed.emit()
        self._rebuild_generation_placeholders()

    def add_files(self, files, image_seq_details=None, quiet=False,
                  prevent_image_seq=False, prevent_recent_folder=False):
        # Access translations
        app = get_app()
        settings = app.get_settings()
        _ = app._tr

        # Make sure we're working with a list of files
        if not isinstance(files, (list, tuple)):
            files = [files]
        scroll_to_files = []

        start_count = len(files)
        for count, filepath in enumerate(files):
            (dir_path, filename) = os.path.split(filepath)

            # Check for this path in our existing project data
            new_file = File.get(path=filepath)

            # If this file is already found, exit
            if new_file:
                # Still add the file (to be selected and scrolled to)
                scroll_to_files.append(new_file)
                del new_file
                continue

            try:
                # Inspect the file with a lightweight temporary reader.
                file_data, _media_duration = inspect_media(
                    filepath,
                    max_width=IMPORT_READER_MAX_SIZE,
                    max_height=IMPORT_READER_MAX_SIZE,
                )

                # Determine media type
                file_data["media_type"] = get_media_type(file_data)

                # Check for audio-only files
                if file_data.get("has_audio") and not file_data.get("has_video"):
                    # Audio-only file should match the current project size and FPS
                    project = get_app().project
                    file_data["width"] = project.get("width")
                    file_data["height"] = project.get("height")

                # Save new file to the project data
                new_file = File()
                new_file.data = file_data

                # Is this an image sequence / animation?
                seq_info = None
                if not prevent_image_seq:
                    seq_info = image_seq_details or self.get_image_sequence_details(filepath)

                if seq_info:
                    # Update file with image sequence path & name
                    new_path = seq_info.get("path")

                    # Load image sequence (to determine duration and video_length)
                    new_file.data, media_duration = inspect_media(
                        new_path,
                        max_width=IMPORT_READER_MAX_SIZE,
                        max_height=IMPORT_READER_MAX_SIZE,
                    )
                    if media_duration > 0.0:
                        # Update file details
                        new_file.data["media_type"] = "video"
                        duration = new_file.data["duration"]

                        if seq_info and "fps" in seq_info and "length_multiplier" in seq_info:
                            # Blender Titles specify their fps in seq_info
                            fps_num = seq_info.get("fps", {}).get("num", 25)
                            fps_den = seq_info.get("fps", {}).get("den", 1)
                            log.debug("Image Sequence using specified FPS: %s / %s" % (fps_num, fps_den))
                        else:
                            # Get the project's fps, apply to the image sequence.
                            fps_num = get_app().project.get("fps").get("num", 30)
                            fps_den = get_app().project.get("fps").get("den", 1)
                            log.debug("Image Sequence using project FPS: %s / %s" % (fps_num, fps_den))

                        # Adjust FPS (difference between 25 FPS and actual FPS)
                        duration *= 25.0 / (float(fps_num) / float(fps_den))
                        new_file.data["duration"] = duration
                        new_file.data["fps"] = {"num": fps_num, "den": fps_den}
                        new_file.data["video_timebase"] = {"num": fps_den, "den": fps_num}

                        log.info(f"Imported '{new_path}' as image sequence with '{fps_num}/{fps_den}' FPS "
                                 f"and '{duration}' duration")

                        # Remove any other image sequence files from the list we're processing
                        match_glob = "{}{}.{}".format(seq_info.get("base_name"), '[0-9]*', seq_info.get("extension"))
                        log.debug("Removing files from import list with glob: {}".format(match_glob))
                        for seq_file in glob.iglob(os.path.join(seq_info.get("folder_path"), match_glob)):
                            # Don't remove the current file, or we mess up the for loop
                            if seq_file in files and seq_file != filepath:
                                files.remove(seq_file)
                    else:
                        # Failed to import image sequence
                        log.info(f"Failed to parse image sequence pattern {new_path}, ignoring...")
                        continue

                if not seq_info:
                    # Log our not-an-image-sequence import
                    log.info("Imported media file {}".format(filepath))

                # Save file
                new_file.save()
                scroll_to_files.append(new_file)

                if start_count > 15:
                    message = _("Importing %(count)d / %(total)d") % {
                            "count": count,
                            "total": len(files) - 1
                            }
                    app.window.statusBar.showMessage(message, 15000)

                # Let the event loop run to update the status bar
                get_app().processEvents()
                # Update the recent import path
                if not prevent_recent_folder:
                    settings.setDefaultPath(settings.actionType.IMPORT, dir_path)

            except Exception as ex:
                # Log exception
                log.warning("Failed to import {}: {}".format(filepath, ex))

                if not quiet and start_count == 1:
                    # Show message box to user (if importing a single file)
                    app.window.invalidImage(filename)

        # Reset list of ignored paths
        self.ignore_image_sequence_paths = []

        # Select all new files (clear previous selection)
        self.selection_model.clearSelection()
        last_selected_index = QModelIndex()
        for file_object in scroll_to_files:
            # Get the index of the newly added file in the proxy model
            index = self.proxy_model.get_file_index(file_object.id)
            if index.isValid():
                # Select & scroll to selection
                self.selection_model.select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
                get_app().window.filesView.scrollTo(
                    model_index_sibling_at_column(index, 0),
                    QAbstractItemView.PositionAtCenter,
                )
                last_selected_index = index
        if last_selected_index.isValid():
            # Keep current index aligned with the newly selected file so actions
            # (preview/properties/etc.) resolve to the expected item.
            self.selection_model.setCurrentIndex(last_selected_index, QItemSelectionModel.NoUpdate)

        message = _("Imported %(count)d files") % {"count": len(files) - 1}
        app.window.statusBar.showMessage(message, 3000)

    def get_image_sequence_details(self, file_path):
        """Inspect a file path and determine if this is an image sequence"""

        # Get just the file name
        (dirName, fileName) = os.path.split(file_path)

        # Image sequence imports are one per directory per run
        if dirName in self.ignore_image_sequence_paths:
            return None

        extensions = ["png", "jpg", "jpeg", "tif", "svg"]
        match = re.findall(r"(.*[^\d])?(0*)(\d+)\.(%s)" % "|".join(extensions), fileName, re.I)

        if not match:
            # File name does not match an image sequence
            return None

        # Get the parts of image name
        base_name = match[0][0]
        fixlen = match[0][1] > ""
        number = int(match[0][2])
        digits = len(match[0][1] + match[0][2])
        extension = match[0][3]

        full_base_name = os.path.join(dirName, base_name)

        # Check for images which the file names have the different length
        fixlen = fixlen or not (
            glob.glob("%s%s.%s" % (full_base_name, "[0-9]" * (digits + 1), extension))
            or glob.glob("%s%s.%s" % (full_base_name, "[0-9]" * ((digits - 1) if digits > 1 else 3), extension))
        )

        # Check for previous or next image
        for x in range(max(0, number - 100), min(number + 101, 50000)):
            if x != number and os.path.exists(
               "%s%s.%s" % (full_base_name, str(x).rjust(digits, "0") if fixlen else str(x), extension)):
                break  # found one!
        else:
            # We didn't discover an image sequence
            return None

        # Found a sequence, ignore this path (no matter what the user answers)
        # To avoid issues with overlapping/conflicting sets of files,
        # we only attempt one image sequence match per directory
        log.debug("Ignoring path for image sequence imports: {}".format(dirName))
        self.ignore_image_sequence_paths.append(dirName)

        log.info('Prompt user to import sequence starting from {}'.format(fileName))
        if not get_app().window.promptImageSequence(fileName):
            # User said no, don't import as a sequence
            return None

        # generate file glob pattern (for this image sequence)
        if not fixlen:
            zero_pattern = "%d"
        else:
            zero_pattern = "%%0%sd" % digits
        pattern = "%s%s.%s" % (base_name, zero_pattern, extension)
        new_file_path = os.path.join(dirName, pattern)

        # Yes, import image sequence
        parameters = {
            "folder_path": dirName,
            "base_name": base_name,
            "fixlen": fixlen,
            "digits": digits,
            "extension": extension,
            "pattern": pattern,
            "path": new_file_path
        }
        return parameters

    def process_urls(self, qurl_list, import_quietly=False, prevent_image_seq=False,
                     transaction_id=None):
        """Recursively process QUrls from a QDropEvent"""
        media_paths = []

        # Transaction — use caller's transaction_id when provided so the
        # caller can group file imports with subsequent operations (e.g.
        # clip creation on timeline drop) into a single undo step.
        owns_transaction = transaction_id is None
        if owns_transaction:
            transaction_id = str(uuid.uuid4())
        get_app().updates.transaction_id = transaction_id

        for uri in qurl_list:
            filepath = uri.toLocalFile()
            if not os.path.exists(filepath):
                continue
            if filepath.endswith(".osp") and os.path.isfile(filepath):
                # Auto load project passed as argument
                get_app().window.OpenProjectSignal.emit(filepath)
                return True
            if os.path.isdir(filepath):
                import_quietly = True
                log.info("Recursively importing {}".format(filepath))
                try:
                    for r, _, f in os.walk(filepath):
                        media_paths.extend(
                            [os.path.join(r, p) for p in f])
                except OSError:
                    log.warning("Directory recursion failed", exc_info=1)
            elif os.path.isfile(filepath):
                media_paths.append(filepath)
        if not media_paths:
            if owns_transaction:
                get_app().updates.transaction_id = None
            return
        # Import all new media files
        media_paths.sort()
        log.debug("Importing file list: {}".format(media_paths))
        self.add_files(media_paths, quiet=import_quietly, prevent_image_seq=prevent_image_seq)
        if owns_transaction:
            get_app().updates.transaction_id = None

    def update_file_thumbnail(self, file_id):
        """Update/re-generate the thumbnail of a specific file"""
        file = File.get(id=file_id)
        path, filename = os.path.split(file.data["path"])
        name = file.data.get("name", filename)

        # Refresh thumbnail for updated file
        self.ignore_updates = True
        m = self.model

        if file_id in self.model_ids:
            # Look up stored index to ID column
            id_index = self.model_ids[file_id]
            if not id_index.isValid():
                return

            thumb_source, _, _ = self._thumbnail_source_for_file(file, clear_cache=True)
            thumb_icon = self._icon_from_thumbnail_source(thumb_source)

            # Update thumb for file
            thumb_index = id_index.sibling(id_index.row(), 0)
            item = m.itemFromIndex(thumb_index)
            item.setIcon(thumb_icon)
            item.setText(name)
            item.setToolTip(self._tooltip_for_file(file, name))
            item.setAccessibleText(name)

            # Update display name
            text_index = id_index.sibling(id_index.row(), 1)
            item = m.itemFromIndex(text_index)
            item.setText(name)

            # Emit signal when model is updated
            self.ModelRefreshed.emit()

        self.ignore_updates = False

    def selected_file_ids(self):
        """ Get a list of file IDs for all selected files """
        # Get the indexes for column 5 of all selected rows
        selected = self.selection_model.selectedRows(5)
        ids = []
        for idx in selected:
            file_id = idx.data()
            if not file_id or self._is_generation_placeholder(file_id):
                continue
            ids.append(file_id)
        return ids

    def selected_files(self):
        """ Get a list of File objects representing the current selection """
        files = []
        for id in self.selected_file_ids():
            files.append(File.get(id=id))
        return files

    def current_file_id(self):
        """ Get the file ID of the current files-view item, or the first selection """
        # Prefer selected rows first, since currentIndex can become stale when
        # switching between details/list views with separate selection models.
        selected_rows = self.selection_model.selectedRows(5)
        if selected_rows:
            selected_ids = set()
            for row_index in selected_rows:
                file_id = row_index.data()
                if file_id and not self._is_generation_placeholder(file_id):
                    selected_ids.add(file_id)

            current = self.selection_model.currentIndex()
            if current and current.isValid():
                current_id = current.sibling(current.row(), 5).data()
                if current_id and current_id in selected_ids:
                    return current_id
            for row_index in selected_rows:
                file_id = row_index.data()
                if file_id and not self._is_generation_placeholder(file_id):
                    return file_id

        cur = self.selection_model.currentIndex()
        if cur and cur.isValid():
            file_id = cur.sibling(cur.row(), 5).data()
            if file_id and not self._is_generation_placeholder(file_id):
                return file_id

    def current_file(self):
        """ Get the File object for the current files-view item, or the first selection """
        cur_id = self.current_file_id()
        if cur_id:
            return File.get(id=cur_id)
        else:
            return None

    def value_updated(self, item):
        """ Table cell change event - when tags are updated on a file"""
        if item.column() == 2:
            # Get updated tag value
            tags_value = item.data(0)
            f = self.current_file()
            if f:
                # Save tags to file object
                f.data["tags"] = tags_value
                f.save()

    def _sync_tree_to_list_selection(self, selected, deselected):
        """Sync selection from TreeView (proxy_model) to ListView (list_proxy_model)"""
        if self._syncing_selection:
            return
        self._syncing_selection = True
        try:
            # Map selected indexes from proxy_model to list_proxy_model
            list_selection = QItemSelection()
            first_list_index = QModelIndex()
            for index in self.selection_model.selectedRows(0):
                list_index = self.list_proxy_model.mapFromSource(index)
                if list_index.isValid():
                    list_selection.select(list_index, list_index)
                    if not first_list_index.isValid():
                        first_list_index = list_index
            self.list_selection_model.select(
                list_selection,
                QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
            )
            if first_list_index.isValid():
                self.list_selection_model.setCurrentIndex(first_list_index, QItemSelectionModel.NoUpdate)
        finally:
            self._syncing_selection = False

    def _sync_list_to_tree_selection(self, selected, deselected):
        """Sync selection from ListView (list_proxy_model) to TreeView (proxy_model)"""
        if self._syncing_selection:
            return
        self._syncing_selection = True
        try:
            # Map selected indexes from list_proxy_model to proxy_model
            tree_selection = QItemSelection()
            first_tree_index = QModelIndex()
            for index in self.list_selection_model.selectedRows(0):
                tree_index = self.list_proxy_model.mapToSource(index)
                if tree_index.isValid():
                    tree_selection.select(tree_index, tree_index)
                    if not first_tree_index.isValid():
                        first_tree_index = tree_index
            self.selection_model.select(
                tree_selection,
                QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
            )
            if first_tree_index.isValid():
                self.selection_model.setCurrentIndex(first_tree_index, QItemSelectionModel.NoUpdate)
        finally:
            self._syncing_selection = False

    def __init__(self, *args, generation_queue=None, proxy_service=None):
        self.generation_queue = generation_queue
        self.proxy_service = proxy_service

        # Add self as listener to project data updates
        # (undo/redo, as well as normal actions handled within this class all update the model)
        app = get_app()
        app.updates.add_listener(self)

        # Create standard model
        self.model = QStandardItemModel()
        self.model.setColumnCount(6)
        self.model_ids = {}
        self.ignore_updates = False
        self.ignore_image_sequence_paths = []

        # Create proxy model (for sorting and filtering) - used by TreeView
        self.proxy_model = FileFilterProxyModel(parent=self)
        self.proxy_model.setDynamicSortFilter(True)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseSensitive)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setSortLocaleAware(True)

        # Create single-column proxy for ListView (wraps proxy_model for accessibility)
        self.list_proxy_model = SingleColumnProxyModel()
        self.list_proxy_model.setSourceModel(self.proxy_model)

        # Connect data changed signal
        self.model.itemChanged.connect(self.value_updated)

        # Create selection models for each view
        self.selection_model = QItemSelectionModel(self.proxy_model)
        self.list_selection_model = QItemSelectionModel(self.list_proxy_model)

        # Sync selections between the two selection models
        self._syncing_selection = False
        self.selection_model.selectionChanged.connect(self._sync_tree_to_list_selection)
        self.list_selection_model.selectionChanged.connect(self._sync_list_to_tree_selection)

        # Connect signal
        app.window.FileUpdated.connect(self.update_file_thumbnail)
        app.window.refreshFilesSignal.connect(
            functools.partial(self.update_model, clear=False))
        if self.generation_queue:
            self.generation_queue.file_job_changed.connect(self._refresh_file_generation_display)
            self.generation_queue.queue_changed.connect(self._refresh_all_generation_displays)
            self.generation_queue.job_added.connect(self._on_generation_job_added)
            self.generation_queue.job_updated.connect(self._on_generation_job_updated)
            self.generation_queue.job_finished.connect(self._on_generation_job_finished)
            self.generation_queue.job_removed.connect(self._on_generation_job_removed)
        if self.proxy_service:
            self.proxy_service.file_job_changed.connect(self._refresh_file_generation_display)
            self.proxy_service.queue_changed.connect(self._refresh_all_generation_displays)

        # Call init for superclass QObject
        super(QObject, FilesModel).__init__(self, *args)

        # Attempt to load model testing interface, if requested
        # (will only succeed with Qt 5.11+)
        if info.MODEL_TEST:
            try:
                # Create model tester objects
                from PyQt5.QtTest import QAbstractItemModelTester
                self.model_tests = []
                for m in [self.proxy_model, self.model]:
                    self.model_tests.append(
                        QAbstractItemModelTester(
                            m, QAbstractItemModelTester.FailureReportingMode.Warning)
                    )
                log.info("Enabled {} model tests for emoji data".format(len(self.model_tests)))
            except ImportError:
                pass

    def _is_generation_placeholder(self, file_id):
        return str(file_id or "").startswith(self.PLACEHOLDER_PREFIX)

    def _placeholder_id_for_job(self, job_id):
        return "{}{}".format(self.PLACEHOLDER_PREFIX, str(job_id or ""))

    def _job_id_from_placeholder(self, file_id):
        file_id = str(file_id or "")
        if not self._is_generation_placeholder(file_id):
            return None
        return file_id[len(self.PLACEHOLDER_PREFIX):]

    def _placeholder_row_for_job(self, job_id):
        placeholder_id = self._placeholder_id_for_job(job_id)
        if placeholder_id not in self.model_ids:
            return None
        id_index = self.model_ids[placeholder_id]
        if not id_index.isValid():
            return None
        return id_index.row()

    def _generation_icon_for_job(self, job):
        icon_name = "tool-generate-sparkle.svg"
        try:
            app = get_app()
            window = getattr(app, "window", None)
            generation_service = getattr(window, "generation_service", None)
            if generation_service and isinstance(job, dict):
                template_id = str(job.get("template_id") or "").strip()
                template = generation_service.template_registry.get_template(template_id)
                if template:
                    resolved_icon = generation_service.icon_for_template(template)
                    if resolved_icon:
                        icon_name = resolved_icon
        except Exception:
            pass

        icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", icon_name)
        if os.path.exists(icon_path):
            return QIcon(icon_path)

        emoji_icon_path = os.path.join(info.PATH, "emojis", "color", "svg", "2728.svg")
        if os.path.exists(emoji_icon_path):
            return QIcon(emoji_icon_path)
        return QIcon(":/icons/Humanity/actions/16/media-record.svg")

    def _add_generation_placeholder(self, job_id):
        job = self.generation_queue.get_job(job_id) if self.generation_queue else None
        if not job:
            return
        if job.get("source_file_id"):
            return

        placeholder_id = self._placeholder_id_for_job(job_id)
        if placeholder_id in self.model_ids and self.model_ids[placeholder_id].isValid():
            self._update_generation_placeholder(job_id)
            return

        name = str(job.get("name") or "generation")
        status = str(job.get("status") or "queued")
        progress = int(job.get("progress", 0))
        progress_detail = str(job.get("progress_detail") or "").strip()
        label = name
        if status == "running":
            label = "{} ({}%)".format(name, progress)
            if progress_detail:
                label = "{} [{}]".format(label, progress_detail)
        elif status == "queued":
            label = "{} (Queued)".format(name)
        elif status == "canceling":
            label = "{} (Canceling...)".format(name)

        row = []
        icon = self._generation_icon_for_job(job)
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemNeverHasChildren

        col = QStandardItem(icon, label)
        col.setFlags(flags)
        row.append(col)

        col = QStandardItem(label)
        col.setFlags(flags)
        row.append(col)

        col = QStandardItem("generation")
        col.setFlags(flags)
        row.append(col)

        col = QStandardItem("generation_job")
        col.setFlags(flags)
        row.append(col)

        col = QStandardItem("")
        col.setFlags(flags)
        row.append(col)

        col = QStandardItem(placeholder_id)
        col.setFlags(flags)
        row.append(col)

        self.model.appendRow(row)
        self.model_ids[placeholder_id] = QPersistentModelIndex(row[5].index())
        self.ModelRefreshed.emit()

    def _update_generation_placeholder(self, job_id):
        row = self._placeholder_row_for_job(job_id)
        if row is None:
            self._add_generation_placeholder(job_id)
            return
        job = self.generation_queue.get_job(job_id) if self.generation_queue else None
        if not job:
            return

        name = str(job.get("name") or "generation")
        status = str(job.get("status") or "queued")
        progress = int(job.get("progress", 0))
        progress_detail = str(job.get("progress_detail") or "").strip()
        label = name
        if status == "running":
            label = "{} ({}%)".format(name, progress)
            if progress_detail:
                label = "{} [{}]".format(label, progress_detail)
        elif status == "queued":
            label = "{} (Queued)".format(name)
        elif status == "canceling":
            label = "{} (Canceling...)".format(name)

        self.model.item(row, 0).setIcon(self._generation_icon_for_job(job))
        self.model.item(row, 0).setText(label)
        self.model.item(row, 1).setText(label)
        left = self.model.index(row, 0)
        right = self.model.index(row, 1)
        self.model.dataChanged.emit(left, right, [Qt.DisplayRole, Qt.AccessibleTextRole])
        self.ModelRefreshed.emit()

    def _remove_generation_placeholder(self, job_id):
        placeholder_id = self._placeholder_id_for_job(job_id)
        if placeholder_id not in self.model_ids:
            return
        id_index = self.model_ids.get(placeholder_id)
        if not id_index or not id_index.isValid():
            self.model_ids.pop(placeholder_id, None)
            return
        row = id_index.row()
        self.model.removeRows(row, 1, id_index.parent())
        self.model.submit()
        self.model_ids.pop(placeholder_id, None)
        self.ModelRefreshed.emit()

    def _rebuild_generation_placeholders(self):
        if not self.generation_queue:
            return
        for job in list(self.generation_queue.jobs.values()):
            if job.get("source_file_id"):
                continue
            if job.get("status") in ("completed", "failed", "canceled"):
                self._remove_generation_placeholder(job.get("id"))
            else:
                self._add_generation_placeholder(job.get("id"))

    def _on_generation_job_added(self, job_id, source_file_id):
        if source_file_id:
            return
        self._add_generation_placeholder(job_id)

    def _on_generation_job_updated(self, job_id, status, progress):
        job = self.generation_queue.get_job(job_id) if self.generation_queue else None
        if not job or job.get("source_file_id"):
            return
        if status in ("completed", "failed", "canceled"):
            self._remove_generation_placeholder(job_id)
        else:
            self._update_generation_placeholder(job_id)

    def _on_generation_job_finished(self, job_id, status):
        job = self.generation_queue.get_job(job_id) if self.generation_queue else None
        if not job or job.get("source_file_id"):
            return
        self._remove_generation_placeholder(job_id)

    def _on_generation_job_removed(self, job_id):
        self._remove_generation_placeholder(job_id)

    def _refresh_file_generation_display(self, file_id):
        file_id = str(file_id or "")
        if not file_id:
            return
        if file_id not in self.model_ids:
            return
        id_index = self.model_ids[file_id]
        if not id_index.isValid():
            return
        row = id_index.row()
        file_obj = File.get(id=file_id)
        if file_obj:
            _, filename = os.path.split(file_obj.data["path"])
            name = file_obj.data.get("name", filename)
            self.model.item(row, 0).setToolTip(self._tooltip_for_file(file_obj, name))
        left = self.model.index(row, 0)
        right = self.model.index(row, 0)
        self.model.dataChanged.emit(left, right, [Qt.DisplayRole, Qt.AccessibleTextRole])
        self.ModelRefreshed.emit()

    def _refresh_all_generation_displays(self):
        if self.model.rowCount() < 1:
            return
        left = self.model.index(0, 0)
        right = self.model.index(self.model.rowCount() - 1, 0)
        self.model.dataChanged.emit(left, right, [Qt.DisplayRole, Qt.AccessibleTextRole])
        self.ModelRefreshed.emit()



"""
 @file
 @brief This file creates the QApplication, and displays the main window
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author olivier Girard <eolinwen@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import atexit
import sys
import os
import platform
import traceback
import json

from PyQt5.QtCore import PYQT_VERSION_STR, QT_VERSION_STR, pyqtSlot
from PyQt5.QtWidgets import QApplication, QMessageBox

# Disable sandbox support for QtWebEngine (required on some Linux distros
# for the QtWebEngineWidgets to be rendered, otherwise no timeline is visible).
# https://doc.qt.io/qt-5/qtwebengine-platform-notes.html#sandboxing-support
os.environ["QTWEBENGINE_DISABLE_SANDBOX"] = "1"


def get_app():
    """ Get the current QApplication instance of OpenShot """
    return QApplication.instance()


def get_settings():
    """Get a reference to the app's settings object"""
    return get_app().get_settings()


class StartupError:
    """ Store and later display an error encountered during setup"""
    levels = {
        "warning": QMessageBox.warning,
        "error": QMessageBox.critical,
    }

    def __init__(self, title="", message="", level="warning"):
        """Create an error message object, populated with details"""
        self.title = title
        self.message = message
        self.level = level

    def show(self):
        """Display the stored error message"""
        box_call = self.levels[self.level]
        box_call(None, self.title, self.message)
        if self.level == "error":
            sys.exit()


class OpenShotApp(QApplication):
    """The primary QApplication subclass for OpenShot."""

    def __init__(self, *args, **kwargs):
        self.mode = kwargs.pop("mode", None)
        super().__init__(*args, **kwargs)
        self.args = super().arguments()
        self.errors = []

        try:
            # Import modules
            from classes import info
            from classes.logger import log, reroute_output

            # Log the session's start
            if self.mode != "unittest":
                import time
                log.info("-" * 48)
                log.info(time.asctime().center(48))
                log.info('Starting new session'.center(48))

            log.debug("Command line: %s", self.args)

            from classes import settings, project_data, updates, sentry
            import openshot

            # Re-route stdout and stderr to logger
            if self.mode != "unittest":
                reroute_output()

        except ImportError as ex:
            tb = traceback.format_exc()
            log.error('OpenShotApp::Import Error', exc_info=1)
            self.errors.append(StartupError(
                "Import Error",
                "Module: %(name)s\n\n%(tb)s" % {"name": ex.name, "tb": tb},
                level="error"))
            # Stop launching
            raise
        except Exception:
            log.error('OpenShotApp::Init Error', exc_info=1)
            sys.exit()

        self.info = info

        # Log some basic system info
        self.log = log
        self.show_environment(info, openshot)
        if self.mode != "unittest":
            self.check_libopenshot_version(info, openshot)

        # Init data objects
        self.settings = settings.SettingStore(parent=self)
        self.settings.load()
        self.apply_timeline_backend_preference()
        self.project = project_data.ProjectDataStore()
        self.updates = updates.UpdateManager()
        # It is important that the project is the first listener if the key gets update
        self.updates.add_listener(self.project)
        self.updates.reset()

        # Set location of OpenShot program (for libopenshot)
        openshot.Settings.Instance().PATH_OPENSHOT_INSTALL = info.PATH

        # Set BABL extensions path
        babl_ext_path = os.path.join(info.PATH, "lib", "babl-ext")
        log.info(f"checking babl_ext_path: {babl_ext_path}")
        if os.path.exists(babl_ext_path):
            os.environ["BABL_PATH"] = babl_ext_path
            log.info(f"setting BABL_PATH: {babl_ext_path}")

        # Check to disable sentry
        if self.mode == "unittest" or not self.settings.get('send_metrics'):
            sentry.disable_tracing()

        # Empty window
        self.window = None

        # Instantiate Theme Manager (Singleton)
        from themes.manager import ThemeManager
        self.theme_manager = ThemeManager(self)

    def show_environment(self, info, openshot):
        log = self.log
        try:
            log.info("-" * 48)
            log.info(("OpenShot (version %s)" % info.SETUP['version']).center(48))
            log.info("-" * 48)

            log.info("openshot-qt version: %s" % info.VERSION)
            log.info("libopenshot version: %s" % openshot.OPENSHOT_VERSION_FULL)
            log.info("platform: %s" % platform.platform())
            log.info("processor: %s" % platform.processor())
            log.info("machine: %s" % platform.machine())
            log.info("python version: %s" % platform.python_version())
            log.info("qt5 version: %s" % QT_VERSION_STR)
            log.info("pyqt5 version: %s" % PYQT_VERSION_STR)

            # Look for frozen version info
            version_path = os.path.join(info.PATH, "settings", "version.json")
            if os.path.exists(version_path):
                with open(version_path, "r", encoding="UTF-8") as f:
                    version_info = json.loads(f.read())
                    log.info("Frozen version info from build server:\n%s" %
                             json.dumps(version_info, indent=4, sort_keys=True))

        except Exception:
            log.debug("Error displaying dependency/system details", exc_info=1)

    def check_libopenshot_version(self, info, openshot):
        """Detect minimum libopenshot version"""
        _ = self._tr
        ver = openshot.OPENSHOT_VERSION_FULL
        min_ver = info.MINIMUM_LIBOPENSHOT_VERSION
        if ver >= min_ver:
            return True

        self.errors.append(StartupError(
            _("Wrong Version of libopenshot Detected"),
            _("<b>Version %(minimum_version)s is required</b>, "
              "but %(current_version)s was detected. "
              "Please update libopenshot or download our latest installer.") % {
                "minimum_version": min_ver,
                "current_version": ver,
                },
            level="error",
        ))

    def apply_timeline_backend_preference(self):
        """Select timeline backend based on preferences, unless CLI overrides it."""
        if not hasattr(self, "settings"):
            return

        # CLI flag overrides preference unless left as 'auto'
        if getattr(self.info, "WEB_BACKEND", "auto") != "auto":
            return

        try:
            use_legacy = bool(self.settings.get("legacy-based-timeline"))
        except Exception:
            self.log.debug("Unable to read legacy timeline setting", exc_info=True)
            return

        if not use_legacy:
            self.info.WEB_BACKEND = "qwidget"
            self.log.info("Legacy timeline disabled via preferences; using QWidget backend.")

    def gui(self):
        """
        Initialize GUI and main window.
        :return: bool: True if the GUI has no errors, False if we fail to initialize the GUI
        """
        from classes import language, sentry, logger_libopenshot

        _ = self._tr
        info = self.info
        log = self.log

        # Init translation system
        language.init_language()
        sentry.set_tag("locale", info.CURRENT_LANGUAGE)

        # Test for permission issues (and display message if needed)
        try:
            log.debug("Testing write access to user directory")
            # Create test paths
            TEST_PATH_DIR = os.path.join(info.USER_PATH, 'PERMISSION')
            TEST_PATH_FILE = os.path.join(TEST_PATH_DIR, 'test.osp')
            os.makedirs(TEST_PATH_DIR, exist_ok=True)
            with open(TEST_PATH_FILE, 'w') as f:
                f.write('{}')
                f.flush()
            # Delete test paths
            os.unlink(TEST_PATH_FILE)
            os.rmdir(TEST_PATH_DIR)
        except PermissionError as ex:
            log.error('Failed to create file %s', TEST_PATH_FILE, exc_info=1)
            self.errors.append(StartupError(
                _("Permission Error"),
                _("%(error)s. Please delete <b>%(path)s</b> and launch OpenShot again.") % {
                    "error": str(ex),
                    "path": info.USER_PATH,
                    },
                level="error",
            ))

        # Display any outstanding startup messages
        self.show_errors()

        # Start libopenshot logging thread
        self.logger_libopenshot = logger_libopenshot.LoggerLibOpenShot()
        self.logger_libopenshot.start()

        # Track which dockable window received a context menu
        self.context_menu_object = None

        # Create main window
        from windows.main_window import MainWindow
        log.debug("Creating main interface window")
        self.window = MainWindow()

        # Check for gui launch failures
        if self.mode == "quit":
            self.window.close()
            return False

        # Clear undo/redo history
        self.window.updateStatusChanged(False, False)

        # Connect our exit signals
        self.aboutToQuit.connect(self.cleanup)

        # Show main window
        self.window.show()

        args = self.args
        if len(args) < 2:
            # Recover backup file (this can't happen until after the Main Window has completely loaded)
            self.window.RecoverBackup.emit()
            return True

        log.info('Process command-line arguments: %s', args[1:])

        # Auto load project if passed as argument
        if args[1].endswith(".osp"):
            self.window.OpenProjectSignal.emit(args[1])
            return True

        # Start a new project and auto import any media files
        self.project.load("")
        for arg in args[1:]:
            self.window.filesView.add_file(arg)
        return True

    def settings_load_error(self, filepath=None):
        """Use QMessageBox to warn the user of a settings load issue"""
        _ = self._tr
        self.errors.append(StartupError(
            _("Settings Error"),
            _("Error loading settings file: %(file_path)s. Settings will be reset.") % {
                "file_path": filepath
                },
            level="warning",
        ))

    def get_settings(self):
        if not hasattr(self, "settings"):
            return None
        return self.settings

    def show_errors(self):
        count = len(self.errors)
        if count > 0:
            self.log.warning("Displaying %d startup messages", count)
        while self.errors:
            error = self.errors.pop(0)
            error.show()

    def _tr(self, message):
        return self.translate("", message)

    @pyqtSlot()
    def cleanup(self):
        """aboutToQuit signal handler for application exit"""
        self.log.debug("Saving settings in app.cleanup")
        try:
            self.settings.save()
        except Exception:
            self.log.error("Couldn't save user settings on exit.", exc_info=1)


@atexit.register
def onLogTheEnd():
    """ Log when the primary Qt event loop ends """
    try:
        from classes.logger import log
        import time
        log.info("OpenShot's session ended".center(48))
        log.info(time.asctime().center(48))
        log.info("=" * 48)
    except Exception:
        import logging
        log = logging.getLogger(".")
        log.debug('Failed to write session ended log', exc_info=1)



"""
 @file
 @brief This file contains repeat time keyframe logic (for Time->Repeat menu)
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import copy
import openshot
from PyQt5.QtWidgets import (
    QDialog, QFormLayout, QComboBox, QSpinBox, QDoubleSpinBox,
    QDialogButtonBox, QHBoxLayout
)
from classes.app import get_app

_ = get_app()._tr


class RepeatDialog(QDialog):
    """Simple dialog to collect custom repeat options."""

    def __init__(self, parent=None, pattern="loop", direction=1):
        super().__init__(parent)
        self.setWindowTitle(_("Custom Repeat"))
        layout = QFormLayout(self)

        self.pattern_combo = QComboBox(self)
        self.pattern_combo.addItems([_("Loop"), _("Ping-Pong")])
        self.pattern_combo.setCurrentIndex(0 if pattern == "loop" else 1)
        layout.addRow(_("Pattern"), self.pattern_combo)

        self.direction_combo = QComboBox(self)
        self.direction_combo.addItems([_("Forward"), _("Reverse")])
        self.direction_combo.setCurrentIndex(0 if direction > 0 else 1)
        layout.addRow(_("Direction"), self.direction_combo)

        self.passes_spin = QSpinBox(self)
        self.passes_spin.setRange(2, 500)
        self.passes_spin.setValue(2)
        layout.addRow(_("Passes"), self.passes_spin)

        delay_layout = QHBoxLayout()
        self.delay_spin = QDoubleSpinBox(self)
        self.delay_spin.setRange(0.0, 100000.0)
        self.delay_spin.setDecimals(3)
        self.delay_unit = QComboBox(self)
        self.delay_unit.addItems([_("frames"), _("ms"), _("sec")])
        delay_layout.addWidget(self.delay_spin)
        delay_layout.addWidget(self.delay_unit)
        layout.addRow(_("Delay"), delay_layout)

        self.ramp_spin = QDoubleSpinBox(self)
        self.ramp_spin.setRange(-1000.0, 1000.0)
        self.ramp_spin.setDecimals(3)
        layout.addRow(_("Speed Ramp (%)"), self.ramp_spin)

        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, parent=self
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def get_values(self, fps_float):
        pattern = "loop" if self.pattern_combo.currentIndex() == 0 else "pingpong"
        direction = 1 if self.direction_combo.currentIndex() == 0 else -1
        passes = self.passes_spin.value()
        delay_val = self.delay_spin.value()
        unit = self.delay_unit.currentText()
        if unit == _("frames"):
            delay_frames = int(round(delay_val))
        elif unit == _("ms"):
            delay_frames = int(round((delay_val / 1000.0) * fps_float))
        else:
            delay_frames = int(round(delay_val * fps_float))
        ramp = self.ramp_spin.value() / 100.0
        return pattern, direction, passes, delay_frames, ramp


# Repeat logic


def _normalize_points(points):
    """Normalize points so X starts at 1 while preserving Y values."""
    if not points:
        return []
    pts = sorted(points, key=lambda p: int(round(p.get("co", {}).get("X", 0))))
    x0 = int(round(pts[0]["co"]["X"]))
    out = []
    for p in pts:
        x = int(round(p["co"].get("X", 0))) - x0 + 1
        y = p["co"].get("Y")
        out.append({"co": {"X": x, "Y": y}, "interpolation": p.get("interpolation", openshot.LINEAR)})
    return out


def _normalize_points_to_trim(points, trim_start_frames, trim_span_frames):
    """Normalize points to the trimmed span (1..trim_span), tolerating clip-relative or absolute."""
    if not points or trim_span_frames <= 0:
        return []

    pts = sorted(points, key=lambda p: int(round(p.get("co", {}).get("X", 0))))
    xs = [int(round(p.get("co", {}).get("X", 0))) for p in pts]
    min_x = min(xs)
    max_x = max(xs)

    normalized = []
    for p, x_abs in zip(pts, xs):
        co = p.get("co", {}) if isinstance(p, dict) else {}

        # If the data already looks clip-relative (within trimmed span), keep as-is
        if 1 <= min_x and max_x <= trim_span_frames:
            x_rel = x_abs
        else:
            # Convert absolute project frame to trimmed-relative frame (1-based)
            x_rel = x_abs - trim_start_frames

        x_rel = int(round(x_rel))

        # Drop any keyframes that fall outside the trimmed clip range once shifted.
        if x_rel < 1 or x_rel > trim_span_frames:
            continue

        new_point = copy.deepcopy(p)
        new_point.setdefault("co", {})
        new_point["co"]["X"] = x_rel
        normalized.append(new_point)
    return normalized


def _repeat_curve(points, span_x, dir_sign, passes, delay_frames, ramp, pattern):
    """Repeat normalized points applying ramp, delay, and direction."""
    new_points = []
    base = 0
    dir_local = dir_sign
    for k in range(passes):
        speed = (1 + ramp) ** k
        scale = 1 / abs(speed)
        dur = max(1, int(round(span_x * scale)))
        pts_iter = points if dir_local > 0 else reversed(points)
        for idx, p in enumerate(pts_iter):
            x = int(round(p["co"].get("X", 0))) - 1
            y = p["co"].get("Y")
            nx_off = min(int(round(x * scale)), dur - 1)
            if dir_local > 0:
                nx = base + nx_off + 1
            else:
                nx = base + (dur - nx_off)
            interp = p.get("interpolation", openshot.LINEAR)
            if pattern == "loop" and k > 0 and idx == 0:
                interp = openshot.CONSTANT
            new_points.append({"co": {"X": nx, "Y": y}, "interpolation": interp})
        base += dur
        if k < passes - 1 and delay_frames:
            last_y = new_points[-1]["co"].get("Y")
            new_points.append({"co": {"X": base, "Y": last_y}, "interpolation": openshot.LINEAR})
            new_points.append({"co": {"X": base + delay_frames, "Y": last_y}, "interpolation": openshot.LINEAR})
            base += delay_frames
        if pattern == "pingpong":
            dir_local *= -1
    return new_points, base

def apply_repeat(clip, pattern, start_dir, passes, delay_frames, ramp, fps_float):
    """Apply repeat stamping to a clip."""
    if passes < 2:
        return

    # Convert trim (seconds) to frames (zero-based repeat: always use 1..trim_span).
    # Use duration to avoid off-by-one loss when end is rounded down/up differently.
    trim_start_frames = int(round(float(clip.data.get("start", 0.0)) * fps_float))
    trim_span_frames = max(
        1,
        int(round((float(clip.data.get("end", 0.0)) - float(clip.data.get("start", 0.0))) * fps_float)),
    )
    target_start_y = trim_start_frames + 1
    target_end_y = target_start_y + trim_span_frames
    target_range = max(1, target_end_y - target_start_y)

    # Normalize existing time curve or build linear default
    orig_time = clip.data.get("time", {}).get("Points", [])
    if isinstance(orig_time, list) and len(orig_time) >= 2:
        base_time = _normalize_points(orig_time)
        y_start = int(round(base_time[0]["co"].get("Y", 0)))
        y_end = int(round(base_time[-1]["co"].get("Y", 0)))
        y_range = max(1, y_end - y_start)
        # Rescale Y values so they align with the trimmed region (in frames)
        for p in base_time:
            y_val = p["co"].get("Y", 0)
            p["co"]["Y"] = target_start_y + ((y_val - y_start) * target_range / y_range)
    else:
        base_frames = trim_span_frames
        base_time = [
            {"co": {"X": 1, "Y": target_start_y}, "interpolation": openshot.LINEAR},
            {"co": {"X": base_frames, "Y": target_end_y}, "interpolation": openshot.LINEAR},
        ]
    time_span_x = max(1, int(round(base_time[-1]["co"]["X"])))

    # Rescale any existing time curve X to the trimmed span so repeat stays zero-based
    if time_span_x != trim_span_frames:
        scale = float(trim_span_frames) / float(time_span_x)
        for p in base_time:
            x_val = int(round(p["co"].get("X", 1) * scale))
            p["co"]["X"] = max(1, x_val)
        time_span_x = trim_span_frames

    # Store original data if not already
    if "repeat_cache" not in clip.data:
        cache = {
            "start": clip.data.get("start", 0.0),
            "end": clip.data["end"],
            "duration": clip.data["duration"],
            "properties": {},
        }
        for k, v in clip.data.items():
            if isinstance(v, dict) and isinstance(v.get("Points"), list) and len(v["Points"]) > 1:
                cache["properties"][k] = copy.deepcopy(v)
        clip.data["repeat_cache"] = cache

    dir_sign = 1 if start_dir >= 0 else -1

    # Build time curve based on existing keyframes
    time_points, total_frames = _repeat_curve(
        base_time, time_span_x, dir_sign, passes, delay_frames, ramp, pattern
    )
    clip.data["time"] = {"Points": time_points}

    # Repeat animated properties
    cache = clip.data.get("repeat_cache", {})
    for prop, original in cache.get("properties", {}).items():
        if prop == "time":
            continue
        norm = _normalize_points_to_trim(
            original.get("Points", []),
            trim_start_frames,
            trim_span_frames,
        )
        span = int(round(norm[-1]["co"]["X"])) if norm else 0
        if span:
            # Rescale keyframe X into the trimmed span to keep repeats aligned
            if span != time_span_x:
                scale = float(time_span_x) / float(span)
                first_x = float(norm[0]["co"].get("X", 1))
                for p in norm:
                    orig_x = float(p["co"].get("X", 1))
                    x_val = 1 + (orig_x - first_x) * scale
                    x_val = int(round(x_val))
                    p["co"]["X"] = min(max(1, x_val), time_span_x)
                span = time_span_x

            # Deduplicate any points that collapsed onto the same frame (esp. at start)
            latest_by_x = {}
            for p in norm:
                x_val = int(round(p["co"].get("X", 1)))
                p["co"]["X"] = x_val
                # Prefer the last occurrence (most recently edited) when duplicates collapse
                latest_by_x[x_val] = p
            norm = [latest_by_x[x] for x in sorted(latest_by_x.keys())]

            new_points, used = _repeat_curve(norm, span, dir_sign, passes, delay_frames, ramp, pattern)
            clip.data[prop] = {"Points": new_points}
            total_frames = max(total_frames, used)

    # Update trims to cover the repeated span starting at 0
    new_duration = total_frames / fps_float
    clip.data["start"] = 0.0
    clip.data["end"] = new_duration
    clip.data["duration"] = new_duration


def reset_repeat(clip):
    cache = clip.data.pop("repeat_cache", None)
    if not cache:
        return
    clip.data["start"] = cache.get("start", clip.data.get("start"))
    clip.data["end"] = cache.get("end", clip.data.get("end"))
    clip.data["duration"] = cache.get("duration", clip.data.get("duration"))
    for prop, data in cache.get("properties", {}).items():
        clip.data[prop] = data
    if "time" not in cache.get("properties", {}):
        clip.data["time"] = {"Points": [{"co": {"X": 1, "Y": 1}, "interpolation": openshot.LINEAR}]}
